
import numpy as np
import inspect
import xtrack as xt




from . import xsuite_helper as xh




def track(outdict, turns=10):
    x_part = []
    px_part = []
    line = outdict['line']
    particles = outdict['particles']
    for n in np.arange(turns):
        line.track(particles, num_turns=1)
        x_part.append(particles.x)
        px_part.append(particles.px)

    return np.array(x_part).transpose(), np.array(px_part).transpose()



def match_tune(rfko,targetx):
    ''' one of the only functions that has the class paradigm'''
    twiss_bf = rfko.line.twiss(method='4d')
    if rfko.sim_params['targetx'] is not None:
        if (targetx-rfko.sim_params['targetx'])<1e-5:
            pass
            #logger.info(f"The given value is different from the previous->{rfko.sim_params['targetx']} ")
    else:
        if (targetx-6.320)<1e-5:
            pass
            # logger.info(f"The given value is different from the previous->{6.320} ")

    rfko.line.match(method='4d', n_steps_max=30, vary=[xt.Vary('kf', step=1e-9), xt.Vary('kd', step=1e-9)],
                    targets=[xt.Target('qx', targetx, tol=0.000005),
                             xt.Target('qy', twiss_bf.qy, tol=0.005)])

    # logger.debug('twiss before matching is stored in twiss_prev attribute')
    rfko.twiss_prev = twiss_bf
    rfko.twiss = rfko.line.twiss(method='4d')


def matching(outdict, targetx=None ,extraction=True ,qse=False ,bumpers=False):
    ''' TODO
    - add the possibility to match the chromaticity
    '''
    line = outdict['line']
    knobs = ['k1prbhf', 'k1prbhd', 'k2prmp', 'k2prmpj']

    # Value matched to the measurements TUNES OF BARE MACHINE
    vals = [0.05714293659989464, -0.05719922125376693, 0.010731455338789136, -0.019018243990104253]
    for i in range(len(knobs)):
        line.vars[knobs[i]] = vals[i]

    if qse:
        line.vars['kprqse'] = 0.09088224  # Value from LSA try length
    if bumpers:
        line.vars['kpebsw23'] = 0  # from LSA not --> -0.001570859932
        line.vars['kpebsw57'] = 5.544860897340004E-4  # from LSA
    if extraction:
        ####### EXTRACTION UNITS
        line.vars['kpebsw23'] = 0  # from LSA not --> -0.001570859932
        line.vars['kpebsw57'] = 5.544860897340004E-4  # from LSA
        line.vars['kprqse'] = 0.09088224  # Value from LSA try length
        line.vars['kprxse'] = 1.4193310000000103  # value at the steady from LSA

        if targetx is None:  ### STANDARD QXTARGET TUNE {6.32 }

            line.vars['kf'] =  0.002252965399686161
            line.vars['kd'] = -0.009766502523392236
            twiss = line.twiss(method='4d')
            outdict['twiss'] = twiss
            print(f'--> standard value Qx = {twiss.qx}')
        elif targetx =='Not':
            # This leaves everything as it is after activating the extraction elements without changing anything else
            print(f'--> Horizontal Tune just after activating the exctraction elements')
            outdict['twiss'] = line.twiss(method='4d')

        else: # TUNE WITH THE LOW ENERGY QUADRUPOLES
            print(f'-->the target qx is {targetx}')
            line.match(method='4d', n_steps_max=30, vary=[xt.Vary('kf', step=1e-9), xt.Vary('kd', step=1e-9)],
                       targets=[xt.Target('qx', targetx, tol=0.000005), xt.Target('qy', outdict['twiss'].qy, tol=0.005)])
            twiss = line.twiss(method='4d')
            print(f'-->after match {twiss.qx}')
            outdict['twiss'] = twiss



def insert_monitor(line, at_index, n_part=10, n_turns=500, name='monitor'):
    line.unfreeze()
    monitor = xt.ParticlesMonitor(
        num_particles=n_part,
        start_at_turn=0,
        stop_at_turn=n_turns,
        auto_to_numpy=True)

    line.insert_element(index=at_index, element=monitor, name=name)
    return monitor


def build_part(line, A=None,X=None,PX=None ,x_ax=False, px_ax=False,s=None, ramping_sex=False,give_normalized=False,**args):
    """ Building particles at specified amplitudes or at specified, absolute normalized coordinates """
    if (A is None)&((X is None)&(PX is None)): # Sanity check
        print('------give values for A or X and PX-----')
        return None
    if A is not None:
        if isinstance(A,np.float64):
            size = 1
        else:
            size = len(A)

        if x_ax:
            x_n = A
            px_n = np.zeros(size)
        elif px_ax:
            px_n = A
            x_n = np.zeros(size)
        else:
            px_n = A / np.sqrt(2)
            x_n = A / np.sqrt(2)

    elif(X is not None)&(PX is not None): # Thanks to the sanity check there are no other cases
        #assert len(X)==len(PX),'X and PX must be the same length'
        x_n = X
        px_n = PX
    ## A check for the tracker

    if line.tracker is None:
        line.build_tracker()
    if ramping_sex:
        xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
        setter = xt.MultiSetter(line, xse_names, field='k2')  # multisetter obj
        xse0 = setter.get_values()  # initial/nominal values
        setter.set_values([0., 0., 0.])  # changes of value

    twiss = line.twiss(method='4d')
    betx = twiss.betx[0]
    alfx = twiss.alfx[0]
    xc = twiss.x[0]
    pxc = twiss.px[0]
    x = np.sqrt(betx)*x_n +xc
    px = (px_n-x_n*alfx)/np.sqrt(betx) +pxc
    #px = px_n/np.sqrt(betx)-x*alfx/betx +pxc
    additional_args = {key:val for key,val in args.items() if key in inspect.getfullargspec(line.build_particles).args}
    ## PARTICLE REF ALREADY CREATED WITH THE LINE
    if give_normalized:
        particles = line.build_particles(method='4d', x_norm=x_n, px_norm=px_n,**additional_args)
    else:
        particles = line.build_particles(method='4d', x=x, px=px,**additional_args)

    if ramping_sex:
        setter.set_values(xse0)
    return particles



def octupoles_off(rfko):
    ''' Quite restrictive assumpion that sextupoles are only on multipoles'''
    oct_names_dict = xh.find_multipole_component(rfko.line.to_pandas(), 'octupole')
    ### I know the only octupoles are non zero in the multipoles
    oct_names = oct_names_dict['multipoles']
    setter = xt.MultiSetter(rfko.line, oct_names, field='knl',index=3)  # initialize the setter with the octupoles coordinate
    # val0 = setter.get_values()  # get init values
    setter.set_values(np.zeros(len(oct_names)))  # set to zero
    rfko.twiss = rfko.line.twiss(method='4d')






########### COULD WORK ALSO WITHOUT THESE??
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import os

from matplotlib.cm import ScalarMappable

from .utility import replace_null

from .slowex_helper import Virtual_sextupole_complete
from .xsuite_helper import Normalize_monitor

import inspect

#######################
##########         STATIC PLOT FUNCTIONS
########################


def stable_tri_plot(dq, S=77, figsize=(12, 8), dmu=None, linewidth=3, translation_vector=None,c='r',**args):
    ''' dmu : defines the rotation angle, which for positive values correspond to a clock-wise rotation
    '''
    eps = 6 * np.pi * dq
    # crossing points
    ABx = - 2 / 3 * eps / S
    ABpx = -eps * 2 / (S * np.sqrt(3))
    ACx = - 2 / 3 * eps / S
    ACpx = eps * 2 / (S * np.sqrt(3))
    CBx = (4 / 3) * eps / S
    CBpx = 0

    #more_params = args
    #plot_params = [arg for arg in args if arg in inspect.getfullargspec(plt.plot).args]
    plot_args = {key:val for key,val in args.items() if key in inspect.getfullargspec(plt.plot).args}
    # print(f'eps = {eps}   AB = {ABx},{ABpx}  AC = {ACx},{ACpx}, CB = {CBx},{CBpx}')
    if dmu is not None: # Here there is a bug ost likely because the rotations becomes distorted
        dmu = dmu % (2 * np.pi)
        # dmu must be in rads
        rotation_matrix = np.array([[np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])
        ABx_r = rotation_matrix.dot(np.array([ABx, ABpx]))[0]
        ABpx_r = rotation_matrix.dot(np.array([ABx, ABpx]))[1]
        ACx_r = rotation_matrix.dot(np.array([ACx, ACpx]))[0]
        ACpx_r = rotation_matrix.dot(np.array([ACx, ACpx]))[1]
        CBx_r = rotation_matrix.dot(np.array([CBx, CBpx]))[0]
        CBpx_r = rotation_matrix.dot(np.array([CBx, CBpx]))[1]

    if translation_vector is not None:  # POCO TESTATO
        if dmu is not None:
            ABx_r += translation_vector[0]
            ACx_r += translation_vector[0]
            CBx_r += translation_vector[0]

            ABpx_r += translation_vector[1]
            ACpx_r += translation_vector[1]
            CBpx_r += translation_vector[1]

        else:
            ABx += translation_vector[0]
            ACx += translation_vector[0]
            CBx += translation_vector[0]

            ABpx += translation_vector[1]
            ACpx += translation_vector[1]
            CBpx += translation_vector[1]


    # plt.figure(figsize=(12, 8)) # LEFT OUT FOR PLOTTING ON TOP OF THE RESULT! NECESSARY FOR ADDING IT TO ANOTHER PLOTTING FUNCTION
    if dmu is not None:
        plt.plot([ABx_r, ACx_r], [ABpx_r, ACpx_r], linewidth=linewidth, c=c,**plot_args)
        plt.plot([ABx_r, CBx_r], [ABpx_r, CBpx_r], linewidth=linewidth, c=c,**plot_args)
        plt.plot([CBx_r, ACx_r], [CBpx_r, ACpx_r], linewidth=linewidth, c=c,**plot_args)
    else:
        plt.plot([ABx, ACx], [ABpx, ACpx], linewidth=linewidth, c=c,**plot_args)
        plt.plot([ABx, CBx], [ABpx, CBpx], linewidth=linewidth, c=c,**plot_args)
        plt.plot([CBx, ACx], [CBpx, ACpx], linewidth=linewidth, c=c,**plot_args)
        print('--- custom plt argument under test')

######### THIS ONE DEALS WITH MULTIPLE TRIANGLE PLOTS
def stable_tri_plot2(dq, S=77, figsize=(12, 8), dmu=None, linewidth=3, translation_vectors=None,c='r',colorbar=False):
    eps = 6 * np.pi * dq
    ABx = []
    ABpx = []
    ACx = []
    ACpx = []
    CBx = []
    CBpx = []
    for ind,x in enumerate(eps):
        
        # crossing points
        ABx.append(- 2 / 3 * x / S)
        ABpx.append(-x * 2 / (S * np.sqrt(3)))
    
        ACx.append(- 2 / 3 * x / S)
        ACpx.append(x * 2 / (S * np.sqrt(3)))
    
        CBx.append((4 / 3) * x / S)
        CBpx.append(0)

    # print(f'raggio_before = {np.sqrt(ABx ** 2 + ABpx ** 2)}')

    # print(f'eps = {eps}   AB = {ABx},{ABpx}  AC = {ACx},{ACpx}, CB = {CBx},{CBpx}')
    if dmu is not None:
        dmu = dmu % (2 * np.pi)
        # dmu must be in rads
        rotation_matrix = np.array([[np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])
        ABx_r = []
        ABpx_r = []
        ACx_r = []
        ACpx_r = []
        CBx_r = []
        CBpx_r = []
                
        for i in range(len(dq)):
            ABx_r.append(rotation_matrix.dot(np.array([ABx[i], ABpx[i]]))[0])
            ABpx_r.append(rotation_matrix.dot(np.array([ABx[i], ABpx[i]]))[1])
            ACx_r.append(rotation_matrix.dot(np.array([ACx[i], ACpx[i]]))[0])
            ACpx_r.append(rotation_matrix.dot(np.array([ACx[i], ACpx[i]]))[1])
            CBx_r.append(rotation_matrix.dot(np.array([CBx[i], CBpx[i]]))[0])
            CBpx_r.append(rotation_matrix.dot(np.array([CBx[i], CBpx[i]]))[1])

    if translation_vectors is not None:  # PENSO NON SERVANO CAMBIAMENTI
        if dmu is not None:
            ABx_r += translation_vectors[:,0]
            ACx_r += translation_vectors[:,0]
            CBx_r += translation_vectors[:,0]

            ABpx_r += translation_vectors[:,1]
            ACpx_r += translation_vectors[:,1]
            CBpx_r += translation_vectors[:,1]

        else:
            ABx += translation_vectors[:,0]
            ACx += translation_vectors[:,0]
            CBx += translation_vectors[:,0]

            ABpx += translation_vectors[:,1]
            ACpx += translation_vectors[:,1]
            CBpx += translation_vectors[:,1]

    # print(f'raggio_after = {np.sqrt(ABx ** 2 + ABpx ** 2)}')
    # plt.figure(figsize=(12, 8)) # LEFT OUT FOR PLOTTING ON TOP OF THE RESULT! NECESSARY FOR ADDING IT TO ANOTHER PLOTTING FUNCTION
    if dmu is not None:
        for ind in range(len(dq)):
            plt.plot([ABx_r[ind], ACx_r[ind]], [ABpx_r[ind], ACpx_r[ind]], linewidth=linewidth, c=c[ind])
            plt.plot([ABx_r[ind], CBx_r[ind]], [ABpx_r[ind], CBpx_r[ind]], linewidth=linewidth, c=c[ind])
            plt.plot([CBx_r[ind], ACx_r[ind]], [CBpx_r[ind], ACpx_r[ind]], linewidth=linewidth, c=c[ind])
    else:
        for ind in range(len(dq)):
            plt.plot([ABx[ind], ACx[ind]], [ABpx[ind], ACpx[ind]], linewidth=linewidth, c=c[ind])
            plt.plot([ABx[ind], CBx[ind]], [ABpx[ind], CBpx[ind]], linewidth=linewidth, c=c[ind])
            plt.plot([CBx[ind], ACx[ind]], [CBpx[ind], ACpx[ind]], linewidth=linewidth, c=c[ind])


def plot_stable(out_dict, s=2,**norm_arg):
    twiss_df = out_dict['twiss'].to_pandas()
    monitor = out_dict['monitor']
    S_strength, S_mu = Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[0] * 2 * np.pi - S_mu
    # dmu from the virtual sextupole to know the rotation
    Monitor = Normalize_monitor(out_dict, normalize_particles=False, center=True,**norm_arg)
    color = np.arange(out_dict['params']['n_turns'])
    colors = np.tile(color, (out_dict['params']['n_part'], 1))
    extracted_mask = Monitor.x != 0
    argmin = np.argmin(np.abs(twiss_df.mux * 2 * np.pi - S_mu))
    Xo = twiss_df.iloc[argmin]['x'] / twiss_df.iloc[argmin]['betx']
    ## CLOSED ORBIT CAUSES A CHANGE IN TUNE SHIFT
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0 - S_strength * Xo / (2 * np.pi)  # distortion correction

    plt.scatter(Monitor.x[extracted_mask], Monitor.px[extracted_mask], s=s, c=colors[extracted_mask])
    plt.colorbar(label='Turns')
    # PROBLEMAAAA SEGNO-ORIENTAMENTO TRIANGOLO
    stable_tri_plot(-dq, S_strength, dmu=dmu)


########### PER LA DISPERSIONE, PLOT STATICI

def plot_dispersion(out_dict,npart=10,norm=False,figsize=(10,8),centered=False):
    ''' To fix:
    - centering is not good
    - remove the extracted now using replace_null

    :param out_dict:
    :param npart:
    :param norm:
    :param figsize:
    :param centered:
    :return:
    '''
    monitor = out_dict['monitor']
    twiss_df = out_dict['twiss'].to_pandas()
    #particles = out_dict['particles']
    
    part_id = np.random.randint(0,out_dict['params']['n_part'],npart)
    deltas = monitor.delta[part_id,:]
    # DISPERSION
    dx = twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
    dpx = twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
    # CLODED ORBIT
    xc = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
    pxc = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]
    
    if (norm is False)&(centered is False):
        x = replace_null(monitor.x[part_id,:])
        px = replace_null(monitor.px[part_id,:])
    elif (norm is False)&(centered is True):
        x = replace_null(monitor.x[part_id, :]) - dx * deltas -xc
        px = replace_null(monitor.px[part_id, :]) - dpx * deltas -pxc

    else: # this should be replaced by the normalizing function
        betx = twiss_df[twiss_df.name=='pe.smh57'].betx.iloc[0]
        alfx = twiss_df[twiss_df.name=='pe.smh57'].alfx.iloc[0]
        if centered is False:
            x = replace_null(monitor.x[part_id,:])/np.sqrt(betx)
            px = x*alfx +replace_null(monitor.px[part_id,:])*np.sqrt(betx)
        else:
            x = (replace_null(monitor.x[part_id, :]) - dx * deltas-xc) / np.sqrt(betx)
            px = x * alfx + (replace_null(monitor.px[part_id, :]) - dpx * deltas-pxc) * np.sqrt(betx)

    plt.figure(figsize=figsize)
    plt.scatter(x,px,c=deltas,s=2)
    plt.colorbar(label='Momentum spread')


def plot_dispersion_n(out_dict, npart=10, norm=False, figsize=(10, 8), centered=False):
    ''' To fix:
    - centering is not good
    - remove the extracted now using replace_null

    :param out_dict:
    :param npart:
    :param norm:
    :param figsize:
    :param centered:
    :return:
    '''
    monitor = out_dict['monitor']
    twiss_df = out_dict['twiss'].to_pandas()
    # particles = out_dict['particles']

    part_id = np.random.randint(0, out_dict['params']['n_part'], npart)
    deltas = monitor.delta[part_id, :]
    # DISPERSION
    dx = twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
    dpx = twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
    # CLODED ORBIT
    xc = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
    pxc = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]

    if norm|centered:
        monitor = Normalize_monitor(out_dict,center=centered,normalize_particles=False)

    x = monitor.x[part_id,:]
    px = monitor.px[part_id,:]
    mask0 = x!=0

    #
    # if (norm is False) & (centered is False):
    #     x = replace_null(monitor.x[part_id, :])
    #     px = replace_null(monitor.px[part_id, :])
    # elif (norm is False) & (centered is True):
    #     x = replace_null(monitor.x[part_id, :]) - dx * deltas - xc
    #     px = replace_null(monitor.px[part_id, :]) - dpx * deltas - pxc
    #
    # else:  # this should be replaced by the normalizing function
    #     betx = twiss_df[twiss_df.name == 'pe.smh57'].betx.iloc[0]
    #     alfx = twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
    #     if centered is False:
    #         x = replace_null(monitor.x[part_id, :]) / np.sqrt(betx)
    #         px = x * alfx + replace_null(monitor.px[part_id, :]) * np.sqrt(betx)
    #     else:
    #         x = (replace_null(monitor.x[part_id, :]) - dx * deltas - xc) / np.sqrt(betx)
    #         px = x * alfx + (replace_null(monitor.px[part_id, :]) - dpx * deltas - pxc) * np.sqrt(betx)

    plt.figure(figsize=figsize)
    plt.scatter(x[mask0], px[mask0], c=deltas, s=2)
    plt.colorbar(label='Momentum spread')


#### NON FUNZIONAAA male!!
def plot_stable_disp(out_dict, n_part= 10, s=2,ids=None,ret_ids=False,figsize=(8,6)):
    twiss_df = out_dict['twiss'].to_pandas()
    monitor = out_dict['monitor']
    ###
    S_strength, S_mu = Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[
              0] * 2 * np.pi - S_mu  # dmu from the virtual sextupole to know the rotation
    if ids is None:
        ids = np.random.randint(0,out_dict['params']['n_part'],n_part)
    
    ## remove closed orbit
    xo = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
    pxo = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]
    x = replace_null(monitor.x[ids, :]) # basta se modifico solo queste
    px =replace_null( monitor.px[ids, :])##

  #  extracted_mask = monitor.x[:, at_turn] == 0

    # NORMALIZATION of coordinates
    betx = twiss_df[twiss_df.name == 'pe.smh57'].betx.iloc[0]
    alfx = twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
    X = x / np.sqrt(betx)
    PX = alfx * x / np.sqrt(betx) + np.sqrt(betx) * px
    Xo = xo / np.sqrt(betx)
    PXo = alfx * xo / np.sqrt(betx) + np.sqrt(betx) * pxo

    X_corr = X - Xo
    PX_corr = PX - PXo

    ## CLOSED ORBIT CAUSES A CHANGE IN TUNE SHIFT
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0  + monitor.delta[ids,0]*out_dict['twiss'].dqx - S_strength / (4 * np.pi) * Xo  # distortion correction
    # print(f'dq = {dq}, dmu = {dmu}')
    Dx = monitor.delta[ids, 0] * twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
    Dpx = monitor.delta[ids, 0] * twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
    # NORMALIZE DISPERSION
    Dx = Dx/np.sqrt(betx)
    Dpx = Dx*alfx + Dpx*np.sqrt(betx)
    
    
    # plt.axvline(out_dict['params']['SEPTA_X_MM']/(1000*np.sqrt(betx))-Xo,c='r',linestyle='--')
    # cm = ScalarMappable()
    # colors_tri = cm.to_rgba(monitor.delta[ids, 0])
    # cm2 = ScalarMappable()
    # color_part = cm2.to_rgba(monitor.delta[ids, :])
    # plt.scatter(X_corr, PX_corr, s=s,c=cm2.to_rgba(monitor.delta[ids, :].flatten()))  # funziona solo se appiattisco uno degli array
    # stable_tri_plot2(-dq,S_strength, translation_vectors=np.array([Dx, Dpx]).transpose(),dmu=dmu, c=cm.to_rgba(monitor.delta[ids, 0]),linewidth=1)
    # #plt.colorbar(cm)
    # if ret_ids:
    #     return ids

    cm = ScalarMappable()
    colors_tri = cm.to_rgba(monitor.delta[ids, 0])
    cm2 = ScalarMappable()
    color_part = cm2.to_rgba(monitor.delta[ids, :])
    fig, ax = plt.subplots(figsize=figsize)
    plt.axvline(out_dict['params']['SEPTA_X_MM'] / (1000 * np.sqrt(betx)) - Xo, c='r', linestyle='--')
    ax.scatter(X_corr, PX_corr, s=s,
               c=cm2.to_rgba(monitor.delta[ids, :].flatten()))  # funziona solo se appiattisco uno degli array
    stable_tri_plot2(dq, S_strength, translation_vectors=np.array([Dx, Dpx]).transpose(), dmu=dmu,
                     c=cm.to_rgba(monitor.delta[ids, 0]), linewidth=1, ax=ax)
    plt.colorbar(cm, ax=ax)
    if ret_ids:
        return ids
      
    


##############
#        ANIMATIONS
#############


def anim_ph_space_stable(out_dict, filename=None, message='',step=5):
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss_df = out_dict['line'].twiss(method='4d').to_pandas()
    SEPTA_X_MM = out_dict['params']['SEPTA_X_MM']

    # UTILE PER EVITARE GROSSI BUG CREDO
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    n_turns = out_dict['params']['n_turns']
    df = pd.DataFrame(data)
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=2)

    betx = twiss_df[twiss_df['name'] == 'pe.smh57'].betx.iloc[0]
    alfx = twiss_df[twiss_df['name'] == 'pe.smh57'].alfx.iloc[0]
    x = replace_null(monitor.x)
    px = replace_null(monitor.px)
    deltas = replace_null(monitor.delta)
    ## STABLE TRIANGLE RELATED
    S_strength, S_mu = Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[
              0] * 2 * np.pi - S_mu  # dmu from the virtual sextupole to know the rotation
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion

    ########
    xc = twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
    pxc = twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]
    # REMOVE CLOSED ORBIT and dispersion distortion
    if np.sum(monitor.delta[:, :].flatten()) != 0:
        dx = twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] / np.sqrt(betx)
        dpx = twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0]
        X = (x - xc - deltas * dx) / np.sqrt(betx)
        PX = (x - xc - deltas * dx) * alfx / np.sqrt(betx) + (px - pxc - deltas * dpx) * np.sqrt(
            betx)
    else:
        X = (x - xc) / np.sqrt(betx)
        PX = (x - xc) * alfx / np.sqrt(betx) + (px - pxc) * np.sqrt(
            betx)
    SEPTA_scaled = np.array([SEPTA_X_MM / (np.sqrt(betx) * 1000) - xc / np.sqrt(betx)])
    # dq = dq0 - S_strength *xc/ (4 * np.pi*np.sqrt(betx))-twiss_df[twiss_df.name=='pe.smh57'].dx*deltas   # distortion correction
    dq = dq0 - S_strength * xc / (4 * np.pi * np.sqrt(betx))  # distortion correction

    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame%50==0:
            print(f'frame {frame}')
        ax.clear()
        # position of EXTRACTED particles
        x_extr = df[df['at_turn'] <= at_turn]['x']  #
        px_extr = df[df['at_turn'] <= at_turn]['px']  #
        X_extr = (x_extr - xc) / np.sqrt(betx)
        Px_extr = (x_extr - xc) * alfx / np.sqrt(betx) + (px_extr - pxc) * np.sqrt(betx)

        mask = monitor.x[:, at_turn] != 0  # mask on the monitor array, for the extracted particles (at_turn)

        ax.text(0.1, 0.9, f'{len(x_extr)} extracted', transform=ax.transAxes)
        ax.set_title(f'Phase space turn :{frame}/{n_turns}')
        ax.set_xlabel('x')
        ax.set_ylabel('px')
        ax.axvline(SEPTA_scaled, c="r", linestyle="--")

        # XAXES LIMITS

        axes_margin = 0.1
        # print(f'{SEPTA_scaled} shape of x ={x_for_limits.shape}')
        xmin = np.min(np.concatenate((X.flatten(), SEPTA_scaled)))
        xmax = np.max(X)
        ax.set_xlim(xmin - np.abs((xmin-xmax) * axes_margin), xmax + np.abs((xmax-xmin) * axes_margin))
        # Yaxes LIMITS
        ymin = np.min(PX)
        ymax = np.max(PX)
        ax.set_ylim(ymin - np.abs((ymin-ymax) * axes_margin), ymax + np.abs((ymax-ymin) * axes_margin))

        # ax.scatter(X_extr, PX_extr, c='g', s=2)
        ax.scatter(X[mask, at_turn], PX[mask, at_turn], c='b', s=2)
        ax.scatter(X_extr, Px_extr, c='r', s=2)
        stable_tri_plot(dq, S_strength, dmu=dmu)

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'part{out_dict["params"]["n_part"]}_DPP{out_dict["params"]["DPP_FACTOR"]}_Volt{out_dict["params"]["volt"]}'

    
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/ph_space_stable_tri_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./gif_wes/ph_space_stable_tri_{message}_{filename}.gif', dpi=200)

    return None

######### ANIMATION DISPERSION

def anim_ph_space_disp(out_dict, filename=None, message='',dispersion=False, step=5):
    '''  MAKE ANIMATION OF THE PHASE SPACE AT SEPTA NON-NORMALIZED,
    if dispersion == True it also plots the phase space shifted for the dispersional effect
    '''
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    SEPTA_X_MM = out_dict['params']['SEPTA_X_MM']  # funzionerà?

    # UTILE PER EVITARE GROSSI BUG CREDO
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    n_turns = out_dict['params']['n_turns']
    df = pd.DataFrame(data)
    fig, ax = plt.subplots(figsize=(8, 6))
    scat = ax.scatter([], [], s=2)
    cmap = ScalarMappable()
    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame%50==0:
            print(f'frame {frame}')
        ax.clear()
        # position of EXTRACTED particles
        x_extr = df[df['at_turn'] <= at_turn]['x']  #
        px_extr = df[df['at_turn'] <= at_turn]['px']  #
        delta_extr = df[df['at_turn']<=at_turn]['delta']
        mask = monitor.x[:, at_turn] != 0  # mask on the monitor array, for the extracted particles (at_turn)
        ## particles position
        x_pres = monitor.x[mask, at_turn]
        px_pres = monitor.px[mask, at_turn]

        # TOLGO LA DISPERSIONE
        if dispersion:
            twiss_df = out_dict['line'].twiss(method='4d').to_pandas()
            x_pres_nodisp = x_pres - monitor.delta[mask, at_turn] * twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
            px_pres_nodisp = px_pres - monitor.delta[mask, at_turn] * twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
            x_extr_nodisp = x_extr-delta_extr * twiss_df[twiss_df.name == 'pe.smh57'].dx.iloc[0]
            px_extr_nodisp = px_extr-delta_extr * twiss_df[twiss_df.name == 'pe.smh57'].dpx.iloc[0]
        ax.text(0.1, 0.9, f'{len(x_extr)} extracted', transform=ax.transAxes)
        ax.set_title(f'Phase space turn :{frame}/{n_turns}')
        ax.set_xlabel('x')
        ax.set_ylabel('px')
        axmargin = 0.05
        if SEPTA_X_MM is not None:
            ax.axvline(SEPTA_X_MM / 1000, c="r", linestyle="--")

        # XAXES LIMITS
        if SEPTA_X_MM is not None: # NOT GENERAL TO SET ONLY THE MINIMUM RESPECTIVE TO THE SEPTA
            xmin = np.min(np.concatenate((monitor.x.flatten(), np.array([SEPTA_X_MM / 1000]),particles.x.flatten())))
            xmax = np.max(np.concatenate((monitor.x.flatten(), np.array([SEPTA_X_MM / 1000]),particles.x.flatten())))
        else:
            xmin = np.min(monitor.x.flatten())
            xmax = np.max(monitor.x.flatten())
        ax.set_xlim(xmin - np.abs(xmin - xmax) *axmargin, xmax + np.abs(xmax-xmin) * axmargin)
        # YLIMITS
        ymin = np.min(monitor.px.flatten())
        ymax = np.max(monitor.px.flatten())
        ax.set_ylim(ymin - np.abs(ymin-ymax) * axmargin, ymax + np.abs(ymax-ymin) * axmargin)
        
        ax.scatter(x_extr, px_extr, c=delta_extr, s=2)
        if dispersion is False:
            #ax.scatter(x_pres, px_pres, c=monitor.delta[mask, at_turn], s=2, label='dispersion')
            ax.scatter(x_pres, px_pres, c=cmap.to_rgba(monitor.delta[mask, at_turn]), s=2, label='dispersion')
            
        if dispersion:
            ax.scatter(x_pres_nodisp, px_pres_nodisp, c='r', s=1, alpha=0.5, label='no_dispersion')
            ax.scatter(x_extr_nodisp,px_extr_nodisp,c='c',s=1,alpha=0.5)
            ax.legend(loc='lower left')
        
            

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    plt.colorbar(cmap,label='Dispersion',ax=ax)
    if filename is None:

        filename = f'n_part{out_dict["params"]["n_part"]}_QX{out_dict["twiss"].qx}_Volt{out_dict["params"]["volt"]}_duration{out_dict["params"]["duration"]}'

    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/ph_space_disp_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./gif_wes/ph_space_{message}_{filename}.gif', dpi=200)

    return None

def anim_ph_space_dyn(out_dict, filename=None,message='', step=5):
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss_df = out_dict['line'].twiss(method='4d').to_pandas()
    SEPTA_X_MM = out_dict['params']['SEPTA_X_MM']
    if 'dynamic_twiss' not in out_dict.keys():
        print('ERROR not possible to use this plot')
        return None

    # UTILE PER EVITARE GROSSI BUG CREDO
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    n_turns = out_dict['params']['n_turns']
    df = pd.DataFrame(data)
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=2)

    betx = twiss_df[twiss_df['name'] == 'pe.smh57'].betx.iloc[0]
    alfx = twiss_df[twiss_df['name'] == 'pe.smh57'].alfx.iloc[0]

    x = replace_null(monitor.x)
    px = replace_null(monitor.px)
    deltas = replace_null(monitor.delta)
    
    ########
    xc = twiss_df[twiss_df['name'] == 'pe.smh57'].x.iloc[0]
    pxc = twiss_df[twiss_df['name'] == 'pe.smh57'].px.iloc[0]
    SEPTA_scaled = np.array([SEPTA_X_MM / (np.sqrt(betx) * 1000) - xc / np.sqrt(betx)])
    if np.sum(monitor.delta[:, :].flatten()) != 0:
        dx = twiss_df[twiss_df['name'] == 'pe.smh57'].dx.iloc[0] / np.sqrt(betx)
        dpx = twiss_df[twiss_df['name'] == 'pe.smh57'].dpx.iloc[0]
        X = (x - xc - deltas * dx) / np.sqrt(betx)
        PX = (x - xc - deltas * dx) * alfx / np.sqrt(betx) + (px - pxc - deltas * dpx) * np.sqrt(
            betx)
    else:
        X = (x - xc) / np.sqrt(betx)
        PX = (x - xc) * alfx / np.sqrt(betx) + (px - pxc) * np.sqrt(
            betx)
        
    ## STABLE TRIANGLE RELATED
    S_strength, S_mu = Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[
                      0] * 2 * np.pi - S_mu  # dmu from the virtual sextupole to know the rotation
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0 - S_strength * xc / (4 * np.pi * np.sqrt(betx))  # distortion correction

            
    def update(frame):
        # a frame is a turn
        at_turn = frame
        if frame % 50 == 0:
            print(f'frame {frame}')
        ax.clear()
        if frame < out_dict['params']['dynamic_sextupole']:
            ### CONTINUE
            twiss_dyn = out_dict['dynamic_twiss'][f'twiss_{frame+1}']
            line_dyn = out_dict['dynamic_line'][f'line_{frame+1}']
            twiss_df_dyn = twiss_dyn.to_pandas()
            
            S_strength_loc, S_mu_loc = Virtual_sextupole_complete(dict(line=line_dyn,twiss=twiss_dyn))  # compute sextupole strength and mu
            dmu_loc = twiss_df_dyn[twiss_df_dyn.name == 'pe.smh57'].mux.iloc[
                      0] * 2 * np.pi - S_mu_loc  # dmu from the virtual sextupole to know the rotation
            dq0_loc = (twiss_dyn.qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
            dq_loc = dq0_loc - S_strength_loc * xc / (4 * np.pi * np.sqrt(betx))  # distortion correction

            

            
            # dq = dq0 - S_strength *xc/ (4 * np.pi*np.sqrt(betx))-twiss_df[twiss_df.name=='pe.smh57'].dx*deltas   # distortion correction
            
        # position of EXTRACTED particles
        x_extr = df[df['at_turn'] <= at_turn]['x']  #
        px_extr = df[df['at_turn'] <= at_turn]['px']  #
        X_extr = (x_extr - xc) / np.sqrt(betx)
        Px_extr = (x_extr - xc) * alfx / np.sqrt(betx) + (px_extr - pxc) * np.sqrt(betx)

        mask = monitor.x[:, at_turn] != 0  # mask on the monitor array, for the extracted particles (at_turn)

        # X_extr = (x_extr-xc) / np.sqrt(betx)
        # PX_extr = (x_extr-xc) * alfx / np.sqrt(betx) + (px_extr-pxc) / np.sqrt(betx)

        ax.text(0.1, 0.9, f'{len(x_extr)} extracted', transform=ax.transAxes)
        ax.set_title(f'Phase space turn :{frame}/{n_turns}')
        ax.set_xlabel('x')
        ax.set_ylabel('px')
        ax.axvline(SEPTA_scaled, c="r", linestyle="--")

        # XAXES LIMITS

        axes_margin = 0.1
        # print(f'{SEPTA_scaled} shape of x ={x_for_limits.shape}')
        xmin = np.min(np.concatenate((X.flatten(), SEPTA_scaled)))
        xmax = np.max(x)
        ax.set_xlim(xmin - np.abs((xmin - xmax) * axes_margin), xmax + np.abs((xmax - xmin) * axes_margin))

        # Yaxes LIMITS
        ymin = np.min(PX)
        ymax = np.max(PX)
        ax.set_ylim(ymin - np.abs((ymin - ymax) * axes_margin), ymax + np.abs((ymax - ymin) * axes_margin))

        # ax.scatter(X_extr, PX_extr, c='g', s=2)
        ax.scatter(X[mask, at_turn], PX[mask, at_turn], c='b', s=2)
        ax.scatter(X_extr, Px_extr, c='r', s=2)
        if frame < out_dict['params']['dynamic_sextupole']:
            stable_tri_plot(dq_loc, S_strength_loc, dmu=dmu_loc)
        else:
            stable_tri_plot(dq, S_strength, dmu=dmu)

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'n_part{out_dict["params"]["n_part"]}_QX{out_dict["twiss"].qx}_Volt{out_dict["params"]["volt"]}_duration{out_dict["params"]["duration"]}__adiabatic_turns_{out_dict["params"]["dynamic_sextupole"]}'
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/ph_space_stable_dyn_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./gif_wes/ph_space_stable_dyn_{message}_{filename}.gif', dpi=200)

    return None






import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
import os
# from matplotlib.cm import ScalarMappable
from tqdm import tqdm
import time as timing
from scipy.fftpack import rfft, rfftfreq

import xtrack as xt

from . import xsuite_helper as xh
from . import slowex_helper as slwex
from . import plotting2 as plt2
from . import utility
from .machine_class import Rfko

def binning(outdict, turn_binning=5, current=True, plot_stairs=False, duty_factor=True,time_binning=None,
                 duty_binning=None):
    ''' Binning and spill manipulation from the rfko object (or dictionary for now)'''
    particles = outdict['particles']
    revt = outdict['twiss'].T_rev0

    if duty_binning is None:
        duty_binning=turn_binning
    data = {
        'at_turn': particles.at_turn,
        'x': particles.x,
        'px': particles.px,
        'delta': particles.delta,
    }
    df = pd.DataFrame(data)
    # count extracted particles for every turn
    # considering extracted at_turns 0 == index 0
    bincount = np.bincount(df['at_turn'])[:-1]

    binning = []
    duty = []
    time_step = []
    for x in range(len(bincount) // turn_binning + 1):
        binning.append(np.sum(bincount[x * turn_binning : x * turn_binning + turn_binning]))
        time_step.append(x * turn_binning*revt)
            # time_step[n] : is the istant at the start of the time-at-binning
            # from 0----> (nturns-1)*revolution_period

        #### TO CHECK
        if duty_factor:
            mean = np.mean(bincount[x * turn_binning : x * turn_binning + turn_binning])
            mean_sq = mean ** 2
            sq_mean = np.mean( np.array(bincount[x * turn_binning : x * turn_binning + turn_binning]) ** 2)
            duty.append(mean_sq / sq_mean)

    if current:
        # Divide the counts for the time interval of the binning
        binning = np.array(binning) / time_step[1]
        ### for the DUTY FACTOR THIS SHOULD NOT CHANGE

    else:
        binning = np.array(binning)
    if plot_stairs:
        plt.stairs(binning, time_step)
    if duty_factor:
        return time_step, binning, duty
    else:
        return time_step, binning


def binning_data(folder_name,turn_binning=5,time_binning=None,duty_factor=False):
    ''' Binning and manipulation of the spill FROM DATA'''
    part_at_turn = np.load(f'{folder_name}/part_at_turn.npy')

    bincount = np.bincount(part_at_turn)[:-1]
    ## Getting revolution period, not explicitly saved
    # real time simulated
    time = utility.read_from_txt(f'{folder_name}/params.txt','time ')
    n_turns = utility.read_from_txt(f'{folder_name}/params.txt','n_turns ',type_='int')
    revt = time/n_turns

    binning = []
    time_step = []
    duty = []
    ## The last turn will in general consider a differe number of turns
    # another idea : I could jus elimintae it
    decimal = len(bincount) / turn_binning- len(bincount) // turn_binning
    if decimal>0.5:
        n_binnings = int(len(bincount) // turn_binning) + 1
    else:
        n_binnings = int(len(bincount) // turn_binning) + 1

    for x in range(n_binnings ): # plus or minus one should depend on the remainder
        binning.append(np.sum(bincount[x * turn_binning : x * turn_binning + turn_binning]))
        time_step.append(x * turn_binning*revt)

        if duty_factor:
            mean = np.mean(bincount[x * turn_binning : x * turn_binning + turn_binning])
            mean_sq = mean ** 2
            sq_mean = np.mean( np.array(bincount[x * turn_binning : x * turn_binning + turn_binning]) ** 2)
            duty.append(mean_sq / sq_mean)


    return time_step, binning, duty






def _clean_dependencies(dependencies):
    clean_dep = []
    for x in dependencies:
        clean_dep.append(str(x)[6:].strip("']"))
    return clean_dep


def dependencies_quad(outdict, names_dict):
    line_df = outdict['line'].to_pandas()
    line = outdict['line']
    ## possible keys from the find_multipole_component functions 'multipoles', 'combf_knl', 'quad_mult', 'direct_ele', 'combf_sing'
    for key in names_dict.keys():
        full_dep_list = []
        if len(names_dict[key]) != 0:
            names = names_dict[key]
            # select_field = dict(multipoles='knl[1]',combf_knl='knl[1]',quad_mult='knl[1]',direct_ele='k1',combf_sing='k1')
            dep_list = []
            for name in names:
                if key in ['multipole', 'combf_knl', 'quad_mult']:
                    dep_list.append(list(line.element_refs[name].knl[1]._expr._get_dependencies()))
                elif key in ['direct_ele', 'combf_sing']:

                    dep_list.append(list(line.element_refs[name].k1._expr._get_dependencies()))
                else:
                    print('something is strange man')

            flat_dep = [x for sub in dep_list for x in sub]

            full_dep_list.append(list(set(flat_dep)))
    return _clean_dependencies(full_dep_list[0])


def tracking(outdict, f=100, A=0.001,sextupole_turns = None,twiss=False,return_time=False):
    ''' Performs tracking simulating quadrupole's sinusoidal ripples.

    The ripples are introduced through the variable that control the quadrupole strength and NOT the
    Multisetter object it may be slower because it does a for loop every turn
    '''
    # A is a relative amplitude, relative to the actual value of the variable
    if isinstance(outdict,Rfko):
        rfko = outdict
        outdict = rfko.to_dict()
    line = outdict['line']
    particles = outdict['particles']
    line_df = line.to_pandas()
    line_dict = dict()
    twiss_dict = dict()

    #######

    turns = outdict['params']['n_turns']
    # time passed after every turn, its relative because a shift in time is a phase shift
    t_turn = outdict['twiss'].T_rev0 * np.arange(turns)
    if sextupole_turns is not None:
        assert sextupole_turns>0
        #outdict['params']['dynamic_sextupole']=True # This is added to know if the virtual sextupole has to be computed every turn
        outdict['params']['dynamic_sextupole'] = sextupole_turns
        #take the values of sextupoles and set them
        xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
        setter = xt.MultiSetter(line, xse_names, field='k2')  # multisetter obj
        xse0 = setter.get_values()  # initial/nominal values
        ramp = np.linspace(0, 1, sextupole_turns)

    # FIND THE QUADRUPOLE VARS
    quad_names = xh.find_multipole_component(line_df, 'quadrupole') # find quad elems
    # find the variables underlying
    dep_vars = dependencies_quad(outdict, quad_names)
    amp_val = dict()
    for x in dep_vars:
        # initial values
        amp_val[x] = line.vars[x]._get_value()
    time_before = timing.perf_counter()
    it = 0
    # TRACKING
    for _ in tqdm(np.arange(turns)):
        # each turn changes the value of the variable, ad the twiss and new line and track the turn
        ### (A * amp_val[x]) is consideres the amplitude of the sinusoid
        if sextupole_turns is not None:
            if it < sextupole_turns:
                k2_it = xse0 * ramp[it]
                setter.set_values(k2_it)  # changes of sextupoles value

        for x in amp_val.keys():
            # changes of quadrupole-vars values
            line.vars[x] = amp_val[x] + (A * amp_val[x]) * np.sin(2 * np.pi * f * t_turn[it])


        line_dict[f'line_{it}'] = line
        if twiss:
            twiss_dict[f'twiss_{it}'] = line.twiss(method='4d')
        line.track(particles, num_turns=1)
        it += 1
    time_taken = timing.perf_counter() - time_before
    print(f'Seconds elapsed = {time_taken}')
    #print('temporal_twiss and line saved to twiss_dynamics')
    if twiss:
        outdict['twiss_dynamics'] = twiss_dict
    outdict['line_dynamics'] = line_dict
    # RESETTING TO the INITIAL VALUES
    for key in amp_val.keys():
        line.vars[key] = amp_val[key]
    print(f'Total extracted particles = {np.sum(particles.at_turn<turns)}')

    ##à ADD THE PARAMETERS USED TO THE OUTDITCT and RFKO OBJ (if)
    outdict['params']['qr_amp'] = A
    outdict['params']['qr_f'] = f
    if 'rfko' in locals():
        rfko.sim_params['qr_amp']=A
        rfko.sim_params['qr_f'] = f
    if return_time:
        return time_taken

# alternativa senza usare le variabili per vedere se questo funziona senza errori
#### With multisetter
# def quad_ripples_multisetter(outdict, A=0.1, f=100, out=True):
#     line = outdict['line']
#     particles = outdict['particles']
#     line_df = line.to_pandas()
#     twiss_dict = dict()
#     line_dict = dict()
#     t_turn = np.arange(outdict['params']['n_turns']) * outdict['twiss'].T_rev0
#     quad_names = utility.find_multipole_component(line_df, 'quadrupole')
#     comp = [x for x in quad_names.keys() if len(quad_names[x]) != 0]
#     names_knl = [quad_names[x] for x in comp if x in ['multipole', 'combf_knl', 'quad_mult']]  ### Not used for now
#     names_k1 = [quad_names[x] for x in comp if x in ['direct_ele', 'combf_sing']]
#     name_k1_flat = [x for sub in names_k1 for x in sub]
#     setter1 = xt.MultiSetter(line, name_k1_flat, field='k1')
#     initk1 = setter1.get_values()
#
#     it = 0
#     time_before = timing.time()
#     for _ in tqdm(np.arange(outdict['params']['n_turns'])):
#         k1_it = initk1 + initk1 * A * np.sin(np.pi * 2 * f * t_turn[it])
#         setter1.set_values(k1_it)
#         # if out:
#         # twiss_dict[f'twiss_{it + 1}'] = line.twiss(method='4d')
#         # line_dict[f'line_{it + 1}'] = line
#         line.track(particles, num_turns=1)
#         it += 1
#
#     time_taken = timing.time() - time_before
#     print(f"> finished tracking, {time_taken} seconds elapsed")
#     # if out:
#     # outdict['dynamic_twiss'] = twiss_dict
#     # outdict['dynamic_line'] = line_dict


def norm_ripples(outdict, element='pe.smh57'):
    ''' Center on the distortion and the dispersion (particle-wise) orbit and Normalize
    '''
    monitor = outdict['monitor']
    particles = outdict['particles']
    particles_norm = outdict['particles'].copy()
    X = []
    PX = []
    for n in range(outdict['params']['n_turns']):
        twiss = outdict['twiss_dynamics'][f'twiss_{n}']
        twiss_df = twiss.to_pandas()

        betx = twiss_df[twiss_df.name == element].betx.iloc[0]
        alfx = twiss_df[twiss_df.name == element].alfx.iloc[0]
        xc = twiss_df[twiss_df.name == element].x.iloc[0]
        pxc = twiss_df[twiss_df.name == element].px.iloc[0]

        mask_null = monitor.x[:, n] != 0

        # NORMALIZE the turn with the selection given by mask_null
        # np.array(mask) could be unnecessary
        Xturn = (monitor.x[:, n] - np.array(mask_null) * xc) / np.sqrt(betx)
        PXturn = Xturn * alfx + monitor.px[:, n] * np.array(mask_null) * np.sqrt(betx)

        X.append(Xturn)
        PX.append(PXturn)
        particles_norm.x[particles.at_turn == n] = particles_norm.x[particles.at_turn == n] / np.sqrt(betx)
        particles_norm.px[particles.at_turn == n] = particles_norm.x[particles.at_turn == n] * alfx + particles_norm.px[
            particles.at_turn == n] * np.sqrt(betx)
    return (np.array(X).T, np.array(PX).T, particles_norm)


def anim_ph_qd_ripples(out_dict, filename=None, message='', step=5):
    ''' Phase space plot with the 'central' stable trangle
    ---------TODO
    - Tune correction not completed or not correct
    - Sextupole ramping ?
    -
    '''
    particles = out_dict['particles']
    monitor = out_dict['monitor']
    twiss_df = out_dict['line'].twiss(method='4d').to_pandas()
    SEPTA_X_MM = out_dict['params']['SEPTA_X_MM']  ## also this needs to be adjusted actually

    n_turns = out_dict['params']['n_turns']
    fig, ax = plt.subplots(figsize=(12, 8))
    scat = ax.scatter([], [], s=2)

    X, PX, Particles = norm_ripples(out_dict)
    if 'dynamic_sextupole' not in out_dict['params'].keys():
        S_strength, S_mu = slwex.Virtual_sextupole_complete(out_dict)

    def update(frame):
        # a frame is a turn
        if 'dynamic_sextupole' in out_dict['params'].keys():
            S_strength, S_mu = slwex.Virtual_sextupole_complete(dict(line=out_dict['line'],params=out_dict['params'],twiss=out_dict['twiss_dynamics'][f'twiss_{frame}']))
        at_turn = frame
        if frame % 50 == 0:
            print(f'frame {frame}')
        ax.clear()

        twiss = out_dict['twiss_dynamics'][f'twiss_{frame}']
        twiss_df = twiss.to_pandas()
        xc = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
        pxc = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]
        betx = twiss_df[twiss_df.name == 'pe.smh57'].betx.iloc[0]

        X_extr = Particles.x[particles.at_turn <= at_turn]
        PX_extr = Particles.px[particles.at_turn <= at_turn]

        mask = monitor.x[:, at_turn] != 0  # mask on the monitor array, for the extracted particles (at_turn)
        dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[0] - S_mu

        dq0 = (twiss.qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
        # the tune is different for every particle and so the stable triangle, but not including the dispersion,
        # means we consider just the stable for the zero dispersion particles
        dq = dq0 - S_strength * xc / (2 * np.pi * np.sqrt(betx))  # + twiss.dqx*monitor.delta[mask,at_turn]/(2*np.pi)
        ax.text(0.1, 0.9, f'{len(X_extr)} extracted', transform=ax.transAxes)
        ax.set_title(f'Phase space turn :{frame}/{n_turns}')
        ax.set_xlabel('x')
        ax.set_ylabel('px')
        # ax.axvline(SEPTA_scaled, c="r", linestyle="--")

        # XAXES LIMITS
        axes_margin = 0.1

        # xmin = np.min(np.concatenate((X.flatten(), SEPTA_scaled)))
        xmin = np.min(X)

        xmax = np.max(X)
        ax.set_xlim(xmin - np.abs((xmin - xmax) * axes_margin), xmax + np.abs((xmax - xmin) * axes_margin))

        # Yaxes LIMITS
        ymin = np.min(PX)
        ymax = np.max(PX)
        ax.set_ylim(ymin - np.abs((ymin - ymax) * axes_margin), ymax + np.abs((ymax - ymin) * axes_margin))

        ax.scatter(X[mask, at_turn], PX[mask, at_turn], c='b', s=2)
        ax.scatter(X_extr, PX_extr, c='r', s=2)
        plt2.stable_tri_plot(dq, S_strength, dmu=dmu)

        return scat,

    ani = FuncAnimation(fig, update, frames=range(0, n_turns, step), blit=True)
    if filename is None:
        filename = f'n_part{out_dict["params"]["n_part"]}_QX{out_dict["twiss"].qx}_Volt{out_dict["params"]["volt"]}_duration{out_dict["params"]["duration"]}'
    if os.path.isdir('./gif_wes'):
        ani.save(f'./gif_wes/ph_space_stable_quad_ripples_{message}_{filename}.gif', dpi=200)
    else:
        ani.save(f'./gif_wes/ph_space_stable_quad_ripples_{message}_{filename}.gif', dpi=200)

    return None






import numpy as np
import matplotlib.pyplot as plt

from matplotlib.colors import LinearSegmentedColormap

from . import xsuite_helper as xh
from . import phy_helper as ph
from .machine_class import Rfko

def kobay(x, px, dq=0.005, S=77):
    # x and px must be normalized
    H = 3 * np.pi * dq * (x ** 2 + px ** 2) + 1 / 4 * S * (3 * x * px ** 2 - x ** 3)
    return H

def tri_geom(outdict, interpolate=True,sqrt=True,negate_s=True):
    '''TO ADD DISPERSION CORRECTION
    TODO : BUGS 1. works better wit the wrong correction orbit (betx)
        2. it works better without the interpolation
        3. change of sign also is a problem'''
    if isinstance(outdict,Rfko):
        outdict = outdict.to_dict()


    twiss = outdict['twiss']

    twiss_df = twiss.to_pandas()
    S_str, S_mu = Virtual_sextupole_complete(outdict)  # compute sextupole strength and mu
    if negate_s:
        S_str = - S_str
    argmin = np.argmin(np.abs(twiss_df.mux * 2 * np.pi - S_mu))
    if interpolate:

        betx_old = twiss_df.iloc[argmin]['betx']
        betx = np.interp(S_mu,twiss.mux*2*np.pi,twiss.betx)
        if sqrt:
            Xo_old = twiss_df.iloc[argmin]['x'] / np.sqrt(betx_old)
            Xo = np.interp(S_mu, twiss.mux * 2 * np.pi, twiss.x) / np.sqrt(betx)
        else:
            Xo_old = twiss_df.iloc[argmin]['x'] / betx_old
            Xo = np.interp(S_mu,twiss.mux*2*np.pi,twiss.x)/betx

    else:
        if sqrt:
            betx =  twiss_df.iloc[argmin]['betx']
            Xo = twiss_df.iloc[argmin]['x'] / np.sqrt(betx) # There should be the sqrt on betx, a ctually!
        else:
            Xo = twiss_df.iloc[argmin]['x'] / twiss_df.iloc[argmin]['betx'] # There should be the sqrt on betx, actually!

    # tune
    dq0 = (twiss.qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0 -S_str * Xo / (2 * np.pi)  # distortion correction--- SIGN CHANGE BECAUSE OF THE DIFFERENCEs IN CONVENTIONS
    if S_str==0:
        Hsep = 0
        ampl_sep=0
    else:
        Hsep = ((4*np.pi*dq)**3) /( S_str**2)
        ampl_sep = np.sqrt(48 * np.pi * np.sqrt(3)) * np.abs(dq / S_str)
    return dict(stable_amp=ampl_sep, tune_distance=dq,S=S_str,S_mu=S_mu,Hsep=Hsep)



def Virtual_sextupole(out_dict, plot=False, figsize=(12, 8)):
    if isinstance(out_dict,Rfko):
        out_dict = out_dict.to_dict()
    # take al sextupoles
    line_df = out_dict['line'].to_pandas()
    twiss_df = out_dict['twiss'].to_pandas()
    Sex = line_df[line_df['element_type'] == 'Sextupole'].copy()
    # initialize a list
    S_list = []
    mux_list = []
    # iter sextupole to extract the parameters
    for x in Sex['name']:

        beta = twiss_df[twiss_df['name'] == x]['betx'].iloc[0]  # BETA --- versione senza float
        k = Sex[Sex['name'] == x]['element'].iloc[
            0].k2  # .iloc serve ad estrarre il valore senno me lo tralla come serie
        # kn = k/rigidity   # normalize the sextupole strength --------------------- PROVO SENZA NORMALIZZARE
        l = Sex[Sex['name'] == x]['element'].iloc[0].length  # same for .iloc
        S = 1 / 2 * k * l * beta ** (3 / 2)
        S_list.append(S)
        mux = twiss_df[twiss_df['name'] == x].mux.iloc[0] * (
                    2 * np.pi)  # CONVERSIONE IN RADIANTI è necessaria perchè mux è in unità di radianti
        mux_list.append(mux)

    S_list_np = np.array(S_list)
    mux_list_np = np.array(mux_list)

    virtual_strength = np.sqrt(
        np.sum(S_list_np * np.cos(3 * mux_list_np)) ** 2 + np.sum(S_list_np * np.sin(3 * mux_list_np)) ** 2)
    virtual_mux = (1 / 3) * np.arctan(
        np.sum(S_list_np * np.sin(3 * mux_list_np)) / np.sum(S_list_np * np.cos(3 * mux_list_np)))

    # POLAR PLOT, NOT WORKING BY NOW (TOO many components anyway)
    if plot:

        fig, ax = plt.subplots(subplot_kw={'projection': 'polar'}, figsize=figsize)
        for i in range(len(mux_list_np)):
            ax.plot([(3 * mux_list_np[i]) % (2 * np.pi), (3 * mux_list_np[i]) % (2 * np.pi)], [0, S_list_np[i]],
                    color='r', linewidth=1)

        ax.plot([virtual_mux, virtual_mux], [0, virtual_strength], color='g', linewidth=1)
    return virtual_strength, virtual_mux


def Virtual_sextupole_complete(out_dict, plot=False, figsize=(12, 8),integrate_k2=True):
    ''' Only on epossible doubt about the function which is: The component K2 are already considering the length??'''
    # take all the sextupoles componentes along the line, hopefully
    if isinstance(out_dict,Rfko):
        out_dict = out_dict.to_dict()

    line_df = out_dict['line'].to_pandas()
    twiss_df = out_dict['twiss'].to_pandas()

    nz_name_dict = xh.find_multipole_component(line_df, order='Sextupole')

    sex_direct = nz_name_dict['direct']
    sex_mult = nz_name_dict['multipoles']
    sex_combf = nz_name_dict['combf_knl']
    sex_quad_knl = nz_name_dict['quad_mult']

    # initialize a list
    S_list = []
    mux_list = []

    # iter sextupole to extract the parameters
    for x in sex_direct:
        # beta = float(twiss_df[twiss_df['name']==x]['betx'])
        beta = twiss_df[twiss_df['name'] == x]['betx'].iloc[0]
        k = line_df[line_df.name == x].element.iloc[0].k2
        # kn = k/rigidity   # normalize the sextupole strength --------------------- It is already normalized
        l = line_df[line_df.name == x].element.iloc[0].length
        S = 1 / 2 * k * l * beta ** (3 / 2)
        S_list.append(S)
        mux = twiss_df[twiss_df['name'] == x].mux.iloc[0] * (2 * np.pi)
        mux_list.append(mux)

    for li in [sex_mult, sex_combf, sex_quad_knl]:
        for x in li:
            beta = twiss_df[twiss_df['name'] == x]['betx'].iloc[0]
            k = line_df[line_df.name == x].element.iloc[
                0].knl[2]
            S = (1 / 2) * k *  beta ** (3 / 2)
            S_list.append(S)
            mux = twiss_df[twiss_df['name'] == x].mux.iloc[0] * (2 * np.pi)  # conversion in radians, mux is in unit of tune cycles
            mux_list.append(mux)

    S_list_np = np.array(S_list)
    mux_list_np = np.array(mux_list)

    if np.sum(S_list)==0:
        print('virtual sextupole  is zero, there are no sextupoles components along the line')
        return 0,0

    virtual_strength = np.sqrt(np.sum(S_list_np * np.cos(3 * mux_list_np)) ** 2 + np.sum(S_list_np * np.sin(3 * mux_list_np)) ** 2)

    virtual_mux = (1/3)*np.arctan(np.sum(S_list_np * np.sin(3 * mux_list_np))/np.sum(S_list_np * np.cos(3 * mux_list_np)))

    # POLAR PLOT
    if plot:
        plt.figure(figsize=figsize)
        fig,ax = plt.subplots(subplot_kw={'projection': 'polar'},figsize=figsize)
        for i in range(len(mux_list_np)):
            ax.plot([(3 * mux_list_np[i]) % (2 * np.pi),(3 * mux_list_np[i]) % (2 * np.pi)], [0,S_list_np[i]], color='r', linewidth=1)
        ax.plot([virtual_mux,virtual_mux],[0,virtual_strength],color='g',linewidth=1)
    return virtual_strength,virtual_mux




def H_amplitude(outdict):
    ''' Action of tracked particles????'''
    Monitor = xh.Normalize_monitor(outdict,center=True,normalize_particles=False,keep_zeroes=True)
    tri_dict = tri_geom(outdict)
    dq = tri_dict['tune_distance']
    # Temporary look for the opposite sign in S, if it is better or not
    S = - tri_dict['S']
    Hvals=[]
    for x in range(Monitor.x.shape[0]):
        Hvals.append(kobay(Monitor.x[x,:],Monitor.px[x,:],S=S,dq=dq))

    return np.array(Hvals)




def kick_angle(gain=1, Brho=None, pc=None, beta=None, charge=None):
    ''' Personal Version of the calculations
    - Errors1 corr : include the charge molteplicity in the electric deflection
    - Error2 corr : calculating the current it add the sqrt(2) factor again

    '''
    # VERSION FOR THE IONS and the energy of the baremachine
    if Brho is None:  # if nothing is given uses the standard energies (Ions 650 Mev/nucl)
        rel_par = ph.energy()
        Brho = rel_par['Brho']  # Tm
        pc = rel_par['pc']  # Gev
        beta = rel_par['beta']
        charge = rel_par['charge']
    # PARAMETER OF TFB
    P = 5e3  # W, TFB peak power / electrode
    Z = 100  # Ohm, TFB impedance / electrode
    L = 935e-3  # m, TFB length
    r = 70e-3  # m, TFB separation

    ## Constants
    mu0 = 4 * np.pi * (10 ** -7)  # H/m, vacuum permeability

    # Electric Field
    Vp = np.sqrt(P * Z * 2)  # peak voltage
    V = Vp * gain
    Efield = V / r  # adjusted for gain

    # Magnetic Field
    I = V / Z  # current corrected? CORR2
    # I = np.sqrt((V ** 2 / Z) / Z * 2)  # current
    Hfield = (2 * I) / (2 * np.pi * r)  # adjusted for gain
    Bfield = Hfield * mu0

    # ANGLE
    theta_M = Bfield * L / (Brho)  # easiest version
    theta_E = np.arctan(charge * Efield * L / (pc * beta * 1e9))  ###  cleaner version CORR1
    theta = theta_E + theta_M


    return theta


def kick_angle_(gain=1, Brho=None, pc=None, beta=None, charge=None):
    ''' TO CHECK BEHAVIOURS

    '''
    # VERSION FOR THE IONS and the energy of the baremachine
    if Brho is None:  # if nothing is given uses the standard energies (Ions 650 Mev/nucl)
        rel_par = ph.energy()
        Brho = rel_par['Brho']  # Tm
        pc = rel_par['pc']  # Gev
        beta = rel_par['beta']
        charge = rel_par['charge']
    # PARAMETER OF TFB
    P = 5e3  # W, TFB peak power / electrode
    Z = 100  # Ohm, TFB impedance / electrode
    L = 935e-3  # m, TFB length
    r = 70e-3  # m, TFB separation

    ## Constants
    mu0 = 4 * np.pi * (10 ** -7)  # H/m, vacuum permeability

    # Electric Field
    # Vp = np.sqrt(P * Z * 2)  # peak voltage
    Vp = np.sqrt(P * Z) # 1000 V
    print(Vp)
    V = Vp * gain
    Efield = V / r  # adjusted for gain

    # Magnetic Field
    I = V / Z  # current corrected? CORR2
    # I = np.sqrt((V ** 2 / Z) / Z * 2)  # current
    # Hfield = (2 * I) / (2 * np.pi * r)  # adjusted for gain
    Hfield = I / (2 * np.pi * r)  # adjusted for gain

    Bfield = Hfield * mu0

    # ANGLE
    theta_M = Bfield * L / (Brho)  # easiest version
    theta_E = np.arctan(charge*Efield * L / (pc * beta * 1e9))  ###  cleaner version CORR1
    theta = theta_E + theta_M
    print(f' theta E {theta_E}')
    print(f' theta M {theta_M}')
    return theta


def transform_to_vs(rfko,monitor=None):
    ''' Transform the particles to the virtual sextupole
    TODO ROTATING THE PARTICLES TOO'''
    if monitor is not None:
        if hasattr(rfko,'monitor'):
            mon_temp = rfko.monitor
        rfko.monitor = monitor
    assert hasattr(rfko, 'monitor'), 'Missing monitor'

    Monitor = xh.Normalize_monitor(rfko,center=True,normalize_particles=False,keep_zeroes=True)
    tri_g = tri_geom(rfko)
    assert hasattr(rfko.monitor,'placed_at_element'), rfko.logger.error('monitor has no attribute .placed_at_element')
    dmu  = rfko.twiss.mux[rfko.twiss.name==rfko.monitor.placed_at_element]*2*np.pi-  tri_g['S_mu']
    Xr,PXr = ph.rotate(Monitor.x,Monitor.px,dmu=-dmu[0])
    # the angle is positive when rotating FROM the sextupole, but in this casE it is TO the sextupole
    try: # in case the monitor was given I change and then revert the changes
        rfko.monitor = mon_temp
    except:
        pass
    return Xr,PXr

def action_angle(rfko,monitor=None):
    ''' different version, with the formula from Neidemar's presentation'''
    ## Transform to the VS coordinates
    X,PX = transform_to_vs(rfko,monitor=monitor)
    trig = tri_geom(rfko)
    S = trig['S']
    dq = trig['tune_distance']
    H_turns = kobay(X,PX,S=S,dq=dq) # This is for every turn
    action = X**2+PX**2
    # H = np.mean(H_turns,axis=1)
    ## TO COMPLETE
    rfko.logger.info('Version action-angle Neidemar')
    angle = np.arctan(-PX/X) + np.pi/2
    mask = X>0
    rfko.logger.info('Difference?')
    angle[mask] = angle[mask] + np.pi
    return action,angle











import numpy as np
import matplotlib.pyplot as plt



from . import xsuite_helper as xh
from . import slowex_helper as slwex
from . import xsuite_actions as xa
from . import line_setup
from . import plotting2 as plt2

from . import utility

from . import bare_mach_setup as bms


def plot_stable(out_dict, s=2, figsize=(12, 7), interpolate=False, wrongf=False, center=True):
    twiss_df = out_dict['twiss'].to_pandas()
    monitor = out_dict['monitor']
    S_strength, S_mu = utility.Virtual_sextupole_complete(out_dict)  # compute sextupole strength and mu
    dmu = twiss_df[twiss_df.name == 'pe.smh57'].mux.iloc[0] * 2 * np.pi - S_mu
    # dmu from the virtual sextupole to know the rotation

    Monitor = utility.Normalize_monitor(out_dict, normalize_particles=False, center=center)
    color = np.arange(out_dict['params']['n_turns'])
    colors = np.tile(color, (out_dict['params']['n_part'], 1))

    extracted_mask = Monitor.x != 0
    argmin = np.argmin(np.abs(twiss_df.mux * 2 * np.pi - S_mu))
    # INTERPOLATEE BETX AND X
    if (interpolate) & (wrongf is False):
        xi, idxi = np.unique(twiss_df.mux, return_index=True)
        yi = twiss_df.iloc[idxi].betx
        yix = twiss_df.iloc[idxi].x
        betxinterp = np.interp(S_mu / (2 * np.pi), xi, yi)
        xinterp = np.interp(S_mu / (2 * np.pi), xi, yix)

        print(f'X: nearest={twiss_df.iloc[argmin]["x"]}, interpolated={xinterp}')
        print(f'BETX: nearest= {twiss_df.iloc[argmin]["betx"]}, interpolated= {betxinterp}')
        # INTERPOLATED
        Xo = xinterp / np.sqrt(betxinterp)

    elif wrongf is False:
        ### NEAREST method
        Xo = twiss_df.iloc[argmin]['x'] / np.sqrt(twiss_df.iloc[argmin]['betx'])
    else:
        # WRONG version THAT WORKs BETTER
        Xo = twiss_df.iloc[argmin]['x'] / twiss_df.iloc[argmin]['betx']

        ## CLOSED ORBIT CAUSES A CHANGE IN TUNE SHIFT
    dq0 = (out_dict['twiss'].qx - 6 - 1 / 3)  # normal as without distorsions and dispersion
    dq = dq0 - S_strength * Xo / (2 * np.pi)  # distortion correction

    ####check if DPP_FACTOR NON ZERO ADD CHROMATICITY CORRECTION
    # print(f'The tune is {out_dict["twiss"].qx}')
    # print(f'The tune distance is {dq} and the Sextupole strength {S_strength}')
    # print(f'their ratio (proportional to the area) {dq/S_strength}')

    plt.figure(figsize=figsize)
    plt.scatter(Monitor.x[extracted_mask], Monitor.px[extracted_mask], s=s, c=colors[extracted_mask])
    plt.colorbar(label='Turns')
    if center is False:
        # TRANLATION OF THE TRIANGLE AT THE MONITOR CLOSED ORBIT
        xt = twiss_df[twiss_df.name == 'pe.smh57'].x.iloc[0]
        pxt = twiss_df[twiss_df.name == 'pe.smh57'].px.iloc[0]
        betx = twiss_df[twiss_df.name == 'pe.smh57'].betx.iloc[0]
        alfx = twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
        twiss_df[twiss_df.name == 'pe.smh57'].alfx.iloc[0]
        Xt = xt / np.sqrt(betx)
        Pxt = Xt * alfx + pxt * np.sqrt(betx)
        t_vector = np.array([Xt, Pxt])
        plt2.stable_tri_plot(dq, - S_strength, dmu=dmu, linewidth=1, translation_vector=t_vector)

    else:
        plt2.stable_tri_plot(dq, - S_strength, dmu=dmu, linewidth=1)
    plt.title(
        f'Tune =  {round(out_dict["twiss"].qx, 5)}, ratio tune distance to sextupole = {round(dq / S_strength, 5)}')




def amplitude_growth(outdict, n_parts=3, figsize=(10, 8), plot=True, return_amp=False, exclude_extr=False,label_by='mean'):
    ''' I suppose to take just few particles, if not one
    TO DO :
    - ADD STATISTICS :
        - moving average,
        - std on the moving average?
    '''

    # Normalizing function
    # There was not centering, this shouldn't change many things
    monitor = utility.Normalize_monitor(outdict, keep_zeroes=True, normalize_particles=False,center=True)

    twiss = outdict['twiss']
    if exclude_extr:
        # Takes out the particles that become axtracted at some point
        idnz = []
        for n in range(monitor.x.shape[0]):
            if 0 not in monitor.x[n, :]:
                idnz.append(n)
        ids = np.random.choice(idnz, size=n_parts,replace=False)
    else:
        ids = np.random.choice([n for n in range(outdict['params']['n_part'])],size=n_parts,replace=False)
    amps = np.sqrt(monitor.x[ids, :] ** 2 + monitor.px[ids, :] ** 2)

    freq = 1 / twiss.T_rev0 - twiss.slip_factor * monitor.delta[
        ids, 0] * 1 / twiss.T_rev0  # it doesn't change on a turn basis
    tune = twiss.qx + twiss.dqx * monitor.delta[ids, 0]
    if plot:
        plt.figure(figsize=figsize)
        for i in range(n_parts):
            mean_amp = np.mean(amps[i, :])
            if label_by=='mean':
                plt.plot(amps[i, :], label=f'mean amplitude = {round(mean_amp, 6)}', alpha=0.5)
            elif label_by=='tune':
                plt.plot(amps[i,:],label=f'tune = {round(tune[i],6)}')
            else :
                init_amp = amps[i,0] # This should be the initial amplitude
                plt.plot(amps[i, :], label=f'initial ampl = {round(init_amp, 6)}')
        plt.title(f"Amplitudes per turn with Gain = {outdict['params']['volt']}")
        plt.xlabel('Turns')
        plt.legend()
    if return_amp:
        return amps





####################### PROTYPE AND CKETCH SPACE
def amps_stats(amps,plot=False,figsize=(10,8),xvar='means'):
    ''' PROBLEM IT CONSIDERS ALL THE AMPLITUDE BUT AT SOME POINT IF A PARTICLE IS EXCTRACTED THOSE GOES TO ZERO AND THE
    STATISTICS DOESN'T MAKE SENSE ANYMORE!!

    :param amps:
    :param plot:
    :param figsize:
    :param xvar:
    :return:
    '''
    stds = np.std(amps,axis=1)
    means = np.mean(amps,axis=1)
    rel_std = stds/means
    print('--> statistics of the amplitudes')
    print(f'--> The average std is {np.mean(stds)}, The average mean is {np.mean(means)}')
    if xvar != 'means':
        init_amps = amps[:,0]
    if plot:
        fig, ax = plt.subplots(2, 1, sharex=True, height_ratios=[2, 2], figsize=figsize)
        if xvar=='means':
            ax[0].scatter(means, stds)
            ax[0].set_xlabel('Mean amplitude')
            ax[1].scatter(means, rel_std)
            ax[1].set_xlabel('Mean amplitude')
        else:
            ax[0].scatter(init_amps, stds)
            ax[0].set_xlabel('Init amplitude')
            ax[1].scatter(init_amps, rel_std)
            ax[1].set_xlabel('Init amplitude')

            #labels by init amps
        ax[0].set_ylabel('Standard deviation vs mean')
        ax[1].set_ylabel('Relative standard deviation vs mean')
    return stds,means,rel_std


##############

def quick_routine( n_turns = 1000,n_part = 1,monofreq=True,volt=0,tune_factor=None,targetx=6.32,max_amp= 4e-3,return_all=False):
   if tune_factor is None:
       tune_factor = targetx-6
   if targetx==6.32:
        # To give the standard value already calculated for this value, avoiding to match again
        targetx = None

   lined = line_setup.build_line_baremachine()
   outdict = dict(line=lined['line'], twiss=lined['line'].twiss(method='4d'))
   bms.matching(outdict,targetx=targetx)

   # Monitor
   monitor = xa.insert_monitor(lined['line'], 'pe.smh57', n_turns=n_turns, n_part=n_part)
   slwex.insert_rfko_custom(lined['line'], tune_factor=tune_factor, volt=volt, n_turns=n_turns, monofreq=monofreq)
   outdict['params'] = dict(n_turns=n_turns, n_part=n_part)
   # Generate particles in a range of amplitudes
   A0 = np.linspace(0,max_amp, n_part)
   particles = xa.build_part(lined['line'], A0)
   # particles = bms.build_particles2(lined['line'],n_part=n_part,DPP_FACTOR=0)
   outdict['particles'] = particles
   outdict['monitor'] = monitor
   outdict['params']['volt'] = volt  # for the amplitude growth funcition
   # Just for using fu2.plot_flex
   outdict['params']['SEPTA_X_MM'] = -75 # JUST FOR PLOTTING, THERE's NO SEPTA
   xh.tracking(outdict)
   ### Calculate the amplitudes
   amp = amplitude_growth(outdict, n_parts=n_part, return_amp=True, plot=False)
   ## statistics for the amplitudes
   print('--> basic statistics for the amplitude')
   stds,means,rel_std= amps_stats(amp, plot=True)

   if return_all:
       return dict(outdict=outdict,stats=dict(stds=stds,means=means,rel_std=rel_std))






def amplitude_growth_(outdict, n_parts=3, figsize=(10, 8), return_amp=False, exclude_extr=False):
    '''
    '''

    # I try the normalizing funciton
    monitor = utility.Normalize_monitor(outdict, keep_zeroes=True, normalize_particles=False)
    ids = np.random.randint(len(monitor.x[:, 0]), size=n_parts)
    twiss = outdict['twiss']
    amps = []
    if exclude_extr:
        idnz = []
        for n in range(monitor.x.shape[0]):
            if 0 in monitor.x[n, :]:
                dx_ = np.where(monitor[n,:]==0)[0][0]
            else:
                dx_= None
            amps.append(np.sqrt(monitor.x[n,:dx_]**2+monitor.px[n,:dx_]**2))
        #ids = np.random.choice(np.arange(len(idnz)), size=n_parts)

    #amps = np.sqrt(monitor.x[ids, :idnz] ** 2 + monitor.px[ids, :] ** 2)

    freq = 1 / twiss.T_rev0 - twiss.slip_factor * monitor.delta[
        :, 0] * 1 / twiss.T_rev0  # it doesn't change on a turn basis
    tune = twiss.qx + twiss.dqx * monitor.delta[ids, 0]

    plt.figure(figsize=figsize)
    for i in range(n_parts):
        mean_amp = np.mean(amps[i, :])
        plt.plot(amps[i, :], label=f'mean amplitude = {round(mean_amp, 6)}', alpha=0.5)
        # plt.plot(amps[i,:],label=f'tune = {round(tune[i],4)}')
    plt.title(f"Amplitudes per turn with Gain = {outdict['params']['volt']}")
    plt.xlabel('Turns')
    plt.legend()
    if return_amp:
        return amps



import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import chirp
from scipy.fft import fft, fftfreq
from scipy.signal import find_peaks


def energy(ions=True,E_nuc=.650,E_prot=24):
    ''' All energies in [Gev]
    :param ions: if False considers Protons
    :param E_nuc: Kinetic energy per nucleons
    :return:
    '''
    m_proton_GeV = 0.938272088160
    if ions:
        # Ion properties lead
        A = 208.0
        Z = 82.0
        N = 126.0
        charge = 54.0
        m_neutron_GeV = 0.93957
        m_electron_GeV = 0.000511
        m_u_GeV = 0.9315  # Mean Nucleon mass empirical

        # Mass defect = mass of all constituents - atomic mass
        mass_defect_GeV = Z * m_proton_GeV + N * m_neutron_GeV + (Z - charge) * m_electron_GeV- A * m_u_GeV
        # real mass = mass of actual contituents - mass defect
        E_0 = Z * m_proton_GeV + N * m_neutron_GeV  + (Z - charge) * m_electron_GeV - mass_defect_GeV
        E = E_0 + E_nuc * A
        gamma = E / E_0
        beta = np.sqrt(1 - 1 / gamma ** 2)
        pc = beta * E  ## Gev
        Brho = pc*3.3356/54 #Brho in Tm
        return dict(E_0 = E_0, E_tot = E,beta = beta,gamma = gamma,pc=pc,ions=ions,Brho=Brho,charge=charge)
    else: # Protons
        charge = 1
        E_0 = m_proton_GeV
        E = E_0+ E_prot
        gamma = E/E_0
        beta = np.sqrt(1 - 1 / gamma ** 2)
        pc = beta * E
        Brho = pc * 3.3356 # Brho in Tm
        return dict(E_0=E_0, E_tot=E, beta=beta, gamma=gamma,pc=pc,ions=ions,Brho=Brho,charge=charge)


def twiss_parameters_from_distribution(x, xp):
    """
    This function calculates the Twiss parameters (beta, alpha, gamma) and emittance
    given a distribution of position (x) and angle (xp).

    Args:
        x (numpy.array): A distribution of position
        xp (numpy.array): A distribution of angle

    Returns:
        beta (float): Beta Twiss parameter
        alpha (float): Alpha Twiss parameter
        gamma (float): Gamma Twiss parameter
        epsilon (float): Emittance
    """

    # Calculate averages
    mean_x = np.mean(x)  # average position
    mean_xp = np.mean(xp)  # average angle

    # Calculate second moments
    mean_xx = np.mean(x ** 2)  # second moment of position
    mean_xxp = np.mean(x * xp)  # cross moment
    mean_xpxp = np.mean(xp ** 2)  # second moment of angle

    # Calculate geometric emittance
    # This is done using the formula for geometric emittance, which is the square root of
    # the determinant of the second moment matrix of (x, xp)
    epsilon = np.sqrt(mean_xx * mean_xpxp - mean_xxp ** 2)

    # Calculate Twiss parameters
    beta = mean_xx / epsilon  # Beta is the variance of position divided by emittance
    alpha = -mean_xxp / epsilon  # Alpha is the negative covariance divided by emittance
    gamma = mean_xpxp / epsilon  # Gamma is the variance of angle divided by emittance

    return dict(beta=beta, alpha=alpha, gamma=gamma, epsilon=epsilon)

# Chirps
def generate_chirp(start_freq, end_freq, duration, fs):
    # Time array
    t = np.linspace(0, duration, int(fs * duration), endpoint=False)
    # Generate chirp signal
    chirp_signal = chirp(t, start_freq, duration, end_freq)

    return t, chirp_signal

# Chirp
def generate_chirp_non_linear(start_freq, end_freq, duration, fs):
    # Time array
    t = np.linspace(0, duration, int(fs * duration), endpoint=False)

    # Half duration
    half_duration = duration / 2

    # Generate two halves of chirp signal
    chirp_signal_first_half = chirp(t[:int(fs * half_duration)], start_freq, half_duration, end_freq)
    chirp_signal_second_half = np.zeros_like(t[int(fs * half_duration):])

    # Concatenate the two halves
    chirp_signal = np.concatenate((chirp_signal_first_half, chirp_signal_second_half))

    return t, chirp_signal


def Ft(signal, time_int, figsize=(10, 8),plot=True,peaks=3,ret_freqs=False,fig_ax=None,verbose=False,normalize_dc=False):
    ''' Fourier transform of a signal, taking only positive frequencies'''

    yf = fft(signal)
    xf = fftfreq(len(signal), time_int)

    # Using from zero on and only absolute vals
    xf__ = xf[np.argsort(xf)]
    yf__= np.abs(yf[np.argsort(xf)])
    zero_index = len(xf) // 2
    xf = xf__[zero_index:]
    yf = yf__[zero_index:]
    if normalize_dc:
        yf = yf / np.max(yf[0])
    if plot:
        if fig_ax is not None:
            ax = fig_ax
        else:
            fig, ax = plt.subplots(figsize=figsize)

        ax.plot(xf, yf)
        ax.xaxis.set_tick_params(rotation=50)
        ax.set_xlabel("freq [Hz]")
        ax.set_ylabel("Amplitude [arb.]")

    peakind = []
    cnt = 0
    while len(peakind) == 0:
        # find peaks using a threshold in height
        height = 0.001
        if cnt > 10 :
            peakind, prop = find_peaks(np.abs(yf), height=0)
            if verbose:
                print('--> height set to zero')
            break

        peakind, prop = find_peaks(np.abs(yf), height=height)
        if len(peakind) == 0: # if no peak is found it lowers the height threshold
            height = height - height * 80 / 100
            cnt+=1

    fsort = np.argsort(prop['peak_heights'])
    if verbose:
        print(f'--> the {peaks} heighest peaks are in {xf[peakind[fsort[:-1-peaks:-1]]]}')
        print(f'--> their heights {np.abs(yf[peakind[fsort[:-1-peaks:-1]]])}')
    if plot:
        plt.scatter(xf[peakind[fsort[:-1-peaks:-1]]], np.abs(yf[peakind[fsort[:-1-peaks:-1]]]), c='r')
    if ret_freqs:
        return xf, yf,xf[peakind[fsort[:-1-peaks:-1]]]
    return xf, yf





def rotate(x, px, dmu, center_mean=True):
    ''' Rotate a distribution of points in 2D
    - POSITIVE DMU -> ROTATION CLOCK WISE. x,px must be arrays or matrices
    - the center is the mean of all the values at all the turns, even for different particles
        but this works well if the particles are centered'''

    x_mean = np.mean(x)
    px_mean = np.mean(px)
    if center_mean:
        x = x - x_mean
        px = px - px_mean

    rot_matrix = np.array([[np.cos(dmu), np.sin(dmu)], [-np.sin(dmu), np.cos(dmu)]])
    x_rot = x.copy()
    px_rot = px.copy()
    if x.ndim == 2:
        for i in range(x.shape[0]):
            rotated = rot_matrix.dot([x[i, :], px[i, :]])
            if center_mean:
                x_rot[i, :] = rotated[0] + x_mean
                px_rot[i, :] = rotated[1] + px_mean
            else:
                x_rot[i, :] = rotated[0]
                px_rot[i, :] = rotated[1]
    else:
        rotated = rot_matrix.dot([x, px])
        if center_mean:
            x_rot = rotated[0] + x_mean
            px_rot = rotated[1] + px_mean
        else:
            x_rot = rotated[0]
            px_rot = rotated[1]

    return x_rot, px_rot



import logging
import scipy
import inspect
import os
import xtrack as xt
import xpart as xp
import xobjects as xo


from . import line_setup
from . import xsuite_helper as xh
from . import slowex_helper as slwex
from . import phy_helper as ph
from . import utility
#from bare_mach_setup import matching,insert_rfko,insert_septa_monitor

import copy # used only in check_optics
import matplotlib.pyplot as plt # only for check_optics
import numpy as np # only for check_optics
logger = logging.getLogger(__name__)
handler = logging.StreamHandler()








class Rfko:
    '''TODO : 1. changing the parameters of the simulation after the .matching causes problems related to the _check_parameters method,
        due to the turn overriding by the time. So SUPERCAREFUl especially when changing the number of turns
        2. The double logger proble everytime I runinitialize another instance, this is shit!!'''

    def __init__(self,line_type='bare_machine',ions=True,E_kin=0.650,logging_level=None,build_line=True,line_dir='mad_file',**sim_params):
        ''' By default, it initializes For the energy of the Ions with 650Mev/nucleon
         '''
        self.line_type = line_type
        # BETTER TO KEEP PARAMETER TIED UP IN DICTIONARIES
        # self.ions = ions
        # self.E_kin = E_kin
        self.line_params = dict(E_kin = E_kin,ions = ions)
        self.sim_params = dict(third_frev=1/3,extraction = True,targetx = None,debunched_co=True,n_part=1000,
                    n_turns=1000,DPP_FACTOR=2.35e-3, exn=2.64399538e-06, eyn= 4.252611276914e-07,
                    SEPTA_X_MM=-60,freq_start_ratio=90/100,freq_end_ratio=110/100,gain=0,sampling_freq=1e9,
                    duration=1/1000,time=None,chirp_type='linear',emittance_norm=True,ctx = xo.ContextCpu())
        keys = self.sim_params.keys()
        self.sim_params.update({key: val for key,val in sim_params.items() if key in keys})


        self.sim_type = 'static' # 'ramping_sex', 'quad_ripples', 'quad_ripples_ramping_sex' for now these are the allowe type


        ################################# LOGGING
        if logging_level is None: # This way is definitely not the best
            self.logging_level = 'INFO'
        else:
            self.logging_level = logging_level
        if 'logger' not in self.__dict__:
            self.logger = logger
        if 'handler' not in self.__dict__:
            # To prevent stupid bugs of the logging module
            self.handler = handler
            self.handler.setLevel(self.logging_level)
            self.logger.addHandler(self.handler)

        self.logger.setLevel(self.logging_level)
        self.logger.info(f"logger instance level is set to {self.logging_level}, to change it use the method set_logging")
        self.logger.info('By default line is built at initialization, otherwise use build_line=False')
        if build_line:
            self.build_line(dir=line_dir)

    def set_logging(self,level):
        allowed_string= ['INFO','DEBUG','WARNING','ERROR','CRITICAL']
        allowed_digit = [10,20,30,40,50]
        assert (level in allowed_string)|(level in allowed_digit), 'The level is not recognized'
        self.logger.setLevel(level)
        self.handler.setLevel(level)


    def set_simulation_params(self,interactive=False,param_dict=None):
        nolist = ['N','No','n','no']
        if interactive:
            self.logger.info('Choice, use N for skipping the choice')
            for key,val in self.sim_params.items():
                Type = type(val)
                inp = input(f'Actual : {key}= {val}-->')
                if inp in nolist:
                    continue
                else:
                    # The following CAN THROW AN ERROR IF THE TYPE CANNOT BE TRANSFORMED
                    self.sim_params[key] = Type(inp) # This takes care of teìhe input type
        else:
            assert param_dict is not None, 'Enter a dictionary'
            assert set(self.sim_params.keys())==set(param_dict.keys()), 'The entered dictionary is not compatible'
            self.sim_params = param_dict

    def _check_params(self):
        ''' Check internal consistency of params
        - between params
        - params-against other attibutes and their properties
        TODO - check targetx with twiss, - septa presence with SEPTA_X_MM - n_turns/n_part  with Monitor
            - HENON MAP CASE

        :return:
        '''
        rfko_params = ['duration','sampling_feq','freq_start_ratio','freq_end_ratio','chirp_type']
        henon_exclude = ['SEPTA_X_MM','debunched_co','third_frev']

        assert 'twiss' in self.__dict__, 'No twiss, Matching method has not been called?'
        # The same parameter is interpreted as normalize em or not depending on the value of
        # 'emittance_norm'
        if self.sim_params['emittance_norm']:
            self.sim_params['ex'] = self.sim_params['exn']/(self.line_params['gamma']*self.line_params['beta'])
            self.sim_params['ey'] = self.sim_params['eyn']/(self.line_params['gamma']*self.line_params['beta'])
        else:
            self.logger.info(f'The emittance in the sim_params attribute is considered as non-normalized')
            self.sim_params['ex'] = self.sim_params['exn']
            self.sim_params['ey'] = self.sim_params['eyn']

        trev = self.twiss['T_rev0']
        self.sim_params['revT'] = trev
        if self.sim_params['time'] is not None:
            self.sim_params['n_turns'] =self.sim_params['time']/trev
            self.logger.info(f"Time parameter ovveride number of turns = {self.sim_params['n_turns']}")
        else:
            self.sim_params['time']= self.sim_params['n_turns']*trev

        allowed_sim_type =   ['static' ,'ramping_sex', 'quad_ripples', 'quad_ripples_ramping_sex']
        while self.sim_type not in ['static' ,'ramping_sex', 'quad_ripples', 'quad_ripples_ramping_sex']:
            self.logger.error(f'Simulation type : {self.sim_type} not recognized')
            inp = input(f'type one of the recognized ones : {allowed_sim_type}')
            self.sim_type=inp

        if self.sim_type in ['quad_ripples','quad_ripples_ramping_sex']:
            self.logger.warning('Quadruole ripples not implemented yet, change to static or ramping_sex')

    def build_line(self,dir='mad_file',madout=False,print_info=False,**kwargs):
        ''' Load the bare machine madx files, defining the energy of the beam, creates the line object (attribute line)'''
        assert (self.line_type=='bare_machine')|(self.line_type=='henon_map'), self.logger.error('Only bare_machine and henon_map are viable options')
        if self.line_type=='bare_machine':
            if os.path.exists(dir):
                lined = line_setup.build_line_baremachine(E_nucleons=self.line_params['E_kin'],dir=dir,
                                                          ions=self.line_params['ions'],
                                                          madout=madout,print_info=print_info)
            else:
                path = utility.search_folder('.',dir)
                assert path is not None, f'No {dir} such folder found, can t find the mad file to initialize the line'
                lined = line_setup.build_line_baremachine(E_nucleons=self.line_params['E_kin'], dir=path,
                                                              ions=self.line_params['ions'],
                                                              madout=madout, print_info=print_info)
        if self.line_type=='henon_map':
            arguments = {key:val for key,val in kwargs.items()
                         if key in inspect.getfullargspec(line_setup.henon_map_line).args}
            if len(arguments.keys())==0:
                self.logger.info('Henon map initialized with default values, look at setup values to know which ones')
            lined = line_setup.henon_map_line(**arguments)
            self.henon_params = lined['henon_params']

        self.line = lined['line']
        self.line_params.update(lined['params'])

    def matching(self,tune_knobs0=None,extraction_eles = True,targetx = None,kd=None,kf=None):
        ''' 1. Match bare machine to the measure tunes and chromaticity (using the specified tune_knobs, Default is reccomended)
            2. Activate extraction elements to measured values (option extraction_eles to skip this)
            3. Fine tune the final desired HORIZONTAL tune (to the targetx value)
            -----------------
            TODO: 1. GENERAL KNOBS, 2. QUICK REFERENCE TO THE MEASURED TUNES,
                '''
        assert self.line_type!='henon_map', ('With the Henon map is not possible to match, '
                                             'gives directly the values you need to the method build_line')
        if targetx is not None:
            self.sim_params['targetx'] = targetx
        if tune_knobs0 is None:
            tune_knobs0 = ['k1prbhf', 'k1prbhd', 'k2prmp', 'k2prmpj']
        else:
            self.logger.error('Not implemented for custom knobs yet')

        # Value matched to the measurements TUNES OF BARE MACHINE
        vals = [0.05714293659989464, -0.05719922125376693, 0.010731455338789136, -0.019018243990104253]
        self.logger.debug(f'Use the {tune_knobs0} to match the tunes and chromas values measured in CHIMERA')
        for i in range(len(tune_knobs0)):
            self.line.vars[tune_knobs0[i]] = vals[i]

        if extraction_eles:
            ####### EXTRACTION UNITS
            self.line.vars['kpebsw23'] = 0  # from LSA not --> -0.001570859932
            self.line.vars['kpebsw57'] = 5.544860897340004E-4  # from LSA
            self.line.vars['kprqse'] = 0.09088224  # Value from LSA try length
            self.line.vars['kprxse'] = 1.4193310000000103  # value at the steady from LSA
            # no if
            twiss_bf_leq = self.line.twiss(method='4d')
            self.logger.debug(f'After the activation of extraction elements the tune is = {round(twiss_bf_leq.qx,6)}')

            if self.sim_params['targetx'] is None:  ### STANDARD QXTARGET {6.32 }
                if (kd is not None)|(kf is not None):
                    self.logger.info('params directly given!')
                    self.line.vars['kf'] = kf
                    self.line.vars['kd'] = kd
                    std_target = False
                else:
                    self.line.vars['kf'] = 0.002252965399686161
                    self.line.vars['kd'] = -0.009766502523392236
                    std_target=True

                twiss = self.line.twiss(method='4d')
                self.twiss = twiss
                self.logger.info(f'The tune will be used : {twiss.qx}')
                if std_target:
                    self.sim_params['targetx']=6.32
                else:
                    self.sim_params['targetx'] = round(twiss.qx,5)
            else:
                # IF NOT USING DEFAULT VALUES, or specified values, It actually does the matching
                self.line.match(method='4d', n_steps_max=30, vary=[xt.Vary('kf', step=1e-9), xt.Vary('kd', step=1e-9)],
                           targets=[xt.Target('qx', self.sim_params['targetx'], tol=0.000005),
                                    xt.Target('qy', twiss_bf_leq.qy, tol=0.005)])
                twiss = self.line.twiss(method='4d')
                self.logger.info(f'The target qx is {targetx} --> After match {twiss.qx}')

                self.twiss = twiss
        else:
            if self.sim_params['targetx'] is not None:
                twiss_bf = self.line.twiss(method='4d')
                self.line.match(method='4d', n_steps_max=30, vary=[xt.Vary('kf', step=1e-9), xt.Vary('kd', step=1e-9)],
                                targets=[xt.Target('qx', self.sim_params['targetx'], tol=0.000005),
                                         xt.Target('qy', twiss_bf.qy, tol=0.005)])
            elif (kd is not None)&(kf is not None):
                self.logger.info('params directly given!')
                self.line.vars['kf'] = kf
                self.line.vars['kd'] = kd
            else:
                self.line.vars['kf'] = 0.002252965399686161
                self.line.vars['kd'] = -0.009766502523392236
                self.logger.info('standard value for kf and kd, which for extraction line would give tunex = 6.32')

            self.logger.warning('The extraction element are not activated')
            self.twiss = self.line.twiss(method='4d')
            self.logger.info(f'Tunes : qx={self.twiss.qx}, qy={self.twiss.qy} || Chromaticities : dqx={self.twiss.dqx}, dqy={self.twiss.dqy}')
            self.sim_params['targetx']=self.twiss.qx
        # I COULD CALL A CHECK PARAMS HERE!
        self._check_params()


    def insert_monitor(self,at_index='pe.smh57',name='SEPTA_MONITOR',start_at_turn=0,
                       stop_at_turn=None,num_particles=None,at_s=None,**monitor_args):

        if stop_at_turn is None:
            stop_at_turn = self.sim_params['n_turns']
        if (num_particles is None)&('particle_id_range' not in monitor_args.keys()):
            num_particles= self.sim_params['n_part']
        # probably needed :
        # assert (num-particles is None('particle_id_range' not in monitor_args.keys())
        if at_s is not None:
            at_index = None

        monitor = xt.ParticlesMonitor(
            num_particles=num_particles,start_at_turn=start_at_turn,
            stop_at_turn=stop_at_turn,auto_to_numpy=True,**monitor_args)

        if self.line_type=='henon_map':
            at_s = None
            at_index = 'Henon_sextupole'
        monitor.placed_at_element = at_index # This is not a predefined attribute of the monitor class but it is actualy quite important in practice
        self.line.unfreeze()
        self.line.insert_element(index=at_index, at_s=at_s,element=monitor, name=name)
        self.monitor = monitor

    def insert_rfko(self,name='EXCITER',at_index='pr.kfb97',**rfko_args):
        ############## shortening the notation
        pc = self.line_params['pc']
        beta = self.line_params['beta']
        Brho = self.line_params['Brho']
        charge = self.line_params['charge']
        duration = self.sim_params['duration']
        sampling_freq = self.sim_params['sampling_freq']
        # Kick angle
        max_kick_angle = slwex.kick_angle(self.sim_params['gain'],Brho,pc,beta,charge)
        frev = 1/self.twiss['T_rev0']
        # FREQUANCIES SCANNED
        start_freq = frev * self.sim_params['freq_start_ratio'] * self.sim_params['third_frev']
        end_freq = frev * self.sim_params['freq_end_ratio'] * self.sim_params['third_frev']
        if self.sim_params['chirp_type'] == 'linear':
            t, chirp_signal = ph.generate_chirp(start_freq, end_freq, duration, sampling_freq)
        else:
            t, chirp_signal = ph.generate_chirp_non_linear(start_freq, end_freq, duration, sampling_freq)

        self.line.unfreeze()

        # Exciter config; the duration parameter tells it to repeat the chirp_signal
        # PROBLEM WITH TIME!
        rfko_exciter = xt.Exciter(
            _context=self.sim_params['ctx'],
            samples=chirp_signal,
            sampling_frequency=sampling_freq,
            frev=frev,
            duration=float(self.sim_params['time']),
            start_turn=0,
            knl=[max_kick_angle]
        )

        self.line.insert_element(
            element=rfko_exciter,
            name=name,
            index= at_index
        )
        self.exciter = rfko_exciter


    def insert_septa(self,at_index='pe.smh57',name='SEPTUM'):
        septum = xt.LimitRect(
            min_x=self.sim_params['SEPTA_X_MM'] * 0.001)
        self.line.unfreeze()
        self.line.insert_element(element=septum,name=name,index=at_index)

    def build_particles(self):
        ''' TODO : 1.QUAD_RIPPLES+RAMPING_SEX CASE not considered yet'''
        # EASIER TO TREAT IN A COMPLETELY DIFFERENT WAY THIS CASE
        if self.line_type=='henon_map':
            delta_array = scipy.stats.norm.rvs(loc=0, scale=self.sim_params['DPP_FACTOR'],
                                               size=int(self.sim_params['n_part']))
            # To remove orbit x at the sextupole ?
            sigmas, sigmas_p = xp.generate_2D_gaussian(int(self.sim_params['n_part']))
            if self.line.tracker is None:
                self.line.build_tracker(self.sim_params['ctx'])
            particles = self.line.build_particles(
                method='4d',
                delta=delta_array,
                nemitt_x=self.sim_params['exn'],
                nemitt_y=self.sim_params['eyn'],
                x_norm=sigmas,
                px_norm=sigmas_p,
                y_norm=sigmas,
                py_norm=sigmas_p,
            )
            self.particles = particles
            self.logger.info('DONE')
            return None


        if self.sim_type == 'ramping_sex':
            xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
            setter = xt.MultiSetter(self.line, xse_names, field='k2')  # multisetter obj
            xse0 = setter.get_values()  # initial/nominal values
            setter.set_values([0., 0., 0.])  # changes of value

        twiss_df = self.twiss.to_pandas()
        sigmas, sigmas_p = xp.generate_2D_gaussian(int(self.sim_params['n_part']))

        z_array = scipy.stats.uniform.rvs(loc=0,
                                          scale=self.twiss['circumference'],
                                          size=int(self.sim_params['n_part']))

        # momentum off set will be gauss distributed along zero and DPP_FACTOR will be the standard deviation
        delta_array = scipy.stats.norm.rvs(loc=0, scale=self.sim_params['DPP_FACTOR'], size=int(self.sim_params['n_part']))
        if self.sim_params['debunched_co']:
            # Transport the particles along the line adding the orbit distortion to each point
            sigmas, sigmas_p = xh.nearest_betx(twiss_df, sigmas, sigmas_p, z_array)
        if self.line.tracker is None:
            self.line.build_tracker(self.sim_params['ctx'])
        # match_at_s=z_array, # instead of zeta
        # at_element='ps$start', # because of the previous is needed
        particles = self.line.build_particles(
            method='4d',
            zeta=z_array, # instead of zeta
            delta=delta_array,
            nemitt_x=self.sim_params['exn'],
            nemitt_y=self.sim_params['eyn'],
            x_norm=sigmas,
            px_norm=sigmas_p,
            y_norm=sigmas,
            py_norm=sigmas_p,
        )
        # particles.zeta = z_array
        # particles.delta = delta_array
        if self.sim_type == 'ramping_sex':
            # RESET THE LINE TO THE NOMINAL VALUES
            setter.set_values(xse0)

        self.particles = particles


    def build_particles_custom(self,A=None,X=None,PX=None,xaxis=False,pxaxis=False,s=None,**args):
        '''

        :param A: amplitudes in the normalized phase space of the generated particles
        :param X: X coordinate in Normalized and centered reference system
        :param PX: PX coordinate in Normalized and centered reference system
        :param xaxis: (only if A is given) if true particles lying on xaxis
        :param pxaxis: same as previous
        :param s:
        :param args:
        :return:
        '''
        assert (A is not None) | ((X is not None) & (PX is not None)), self.logger.error('No coordinates specified')  # Sanity check

        if A is not None: # AMPLITUDES GIVEN
            if isinstance(A, np.float64):
                size = 1
            else:
                size = len(A)
            if xaxis:
                x_n = A
                px_n = np.zeros(size)
            elif pxaxis:
                px_n = A
                x_n = np.zeros(size)
            else:
                px_n = A / np.sqrt(2)
                x_n = A / np.sqrt(2)
            if s is not None:
                assert len(s) ==size

        else:

            assert (X is not None)&(PX is not None),'Both coordinates needed X,PX'
            if (not isinstance(X,np.float64))|(not isinstance(PX,np.float64)):
                assert len(X)==len(PX), 'Different sizes not allowed'
            else:
                assert (isinstance(X,np.float64))&(isinstance(PX,np.float64)),\
                    'one is a single number and the other not, pay attention that both are float'
            x_n = X
            px_n = PX

        if self.line.tracker is None:
            self.line.build_tracker()
        if (self.sim_type == 'ramping_sex')&(self.line_type!='henon_map'):
            xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
            setter = xt.MultiSetter(self.line, xse_names, field='k2')  # multisetter obj
            xse0 = setter.get_values()  # initial/nominal values
            setter.set_values([0., 0., 0.])  # changes of value
            self.twiss = self.line.twiss(method='4d')

        if ('twiss' not in self.__dict__)&(self.line_type!='henon_map'):
            self.twiss = self.line.twiss(method='4d')
        # Assumes all the particles are generates at the sape position along the line
        if s is None:
            s = np.zeros(size)
        x,px = xh.DEnormalize_points(self,X=x_n,PX=px_n,at_s=s[0])


        additional_args = {key: val for key, val in args.items() if
                           key in inspect.getfullargspec(self.line.build_particles).args}

        particles = self.line.build_particles(method='4d', x=x, px=px,zeta=s, **additional_args)

        self.particles= particles
        if (self.sim_type == 'ramping_sex')&(self.line_type!='henon_map'):
            setter.set_values(xse0)
            self.twiss = self.line.twiss(method='4d')


    def setup_line(self,rfko=True,monitor=True,septa=True,build_particles=True):
        message = ''
        if ('twiss' not in self.__dict__)&(self.line_type!='henon_map'):
            self.matching()
            message+= 'Matching'
        if self.line_type=='henon_map':
            septa = False
            rfko= False

        if monitor:
            self.insert_monitor()
            message= message+'-monitor'
        if septa:
            self.insert_septa()
            message= message+'-septa'
        else:
            # For avoiding the plot of it in plotting1.plot_flex
            self.sim_params['SEPTA_X_MM']= None
        if rfko:
            self.insert_rfko()
            message= message+'-rfko'
        if self.line.tracker is None:
            self.line.build_tracker(self.sim_params['ctx'])
        if build_particles:
            self.build_particles()
            message= message+'-built_particles'

        self.logger.info(f'{message}  at default positions and names')

    def to_dict(self,monitor=None,particles=None):
        ''' The cases considered are few because the dictionary can always be changed in another moment
        - Monitor: is None or a monitor given
        - Particles: is true or false (this is what it seems I use more
        TODO: automaticaòòy considering if the elemnts are presents'''
        if ('monitor' not in self.__dict__)&(monitor is None):
            monitor = None
        elif monitor is None:
            monitor = self.monitor
        if (particles is None)&('particles' not in self.__dict__):
            particles = None
        elif particles is None:
            particles = self.particles
        if ('twiss' not in self.__dict__)&(self.line_type!='henon_map'):
            self.logger.debug('twiss not present, calculated now')
            self.twiss = self.line.twiss(method='4d')
        if self.line_type=='henon_map':
            outdict = dict(line=self.line, henon_params=self.henon_params, params=self.sim_params, line_params=self.line_params)
        else:
            outdict = dict(line=self.line,twiss=self.twiss,params=self.sim_params,line_params=self.line_params)
        if monitor is not None:
            outdict['monitor'] = monitor
        if particles is not None:
            outdict['particles'] = particles
        return outdict

    # def to_dict__(self, monitor=None, particles=True):
    #     ''' The cases considered are few because the dictionary can always be changed in another moment
    #     - Monitor: is None or a monitor given
    #     - Particles: is true or false (this is what it seems I use more
    #     TODO: automaticaòòy considering if the elemnts are presents'''
    #     if ('monitor' not in self.__dict__) & (monitor is None):
    #         monitor = None
    #     if (particles is None) & ('particles' not in self.__dict__)
    #
    #     elif particles & (monitor is None):
    #
    #         return dict(line=self.line, twiss=self.twiss,
    #                     params=self.sim_params, monitor=self.monitor,
    #                     particles=self.particles)
    #
    #     elif (particles is False) & (monitor is None):
    #         return dict(line=self.line, twiss=self.twiss,
    #                     params=self.sim_params, monitor=self.monitor,
    #                     )
    #     elif (particles is False) & (monitor is not None):
    #         return dict(line=self.line, twiss=self.twiss,
    #                     params=self.sim_params, monitor=monitor)
    #     elif particles & (monitor is not None):
    #         return dict(line=self.line, twiss=self.twiss,
    #                     params=self.sim_params, monitor=monitor,
    #                     particles=self.particles)

    def check_optics(self,plot=True):
        line = self.line.copy()

        line.vars['kpebsw23'] = 0
        line.vars['kpebsw57'] =0
        line.vars['kprqse'] = 0
        line.vars['kprxse'] = 0
        sigmas, sigmas_p = xp.generate_2D_gaussian(int(self.sim_params['n_part']))

        line.unfreeze()
        monitor = xt.ParticlesMonitor(
            num_particles=50,
            start_at_turn=0,
            stop_at_turn=100,
            auto_to_numpy=True)
        line.insert_element(index='ps$start', element=monitor, name='check_opt')
        if line.tracker is None:
            line.build_tracker(self.sim_params['ctx'])

        particles = line.build_particles(
            method='4d',
            x=sigmas,
            px=sigmas_p,
            y=sigmas,
            py=sigmas_p,
        )

        line.track(particles,num_turns=100)
        if self.line_type!='henon_map':
            Monitor = xh.Normalize_monitor(dict(line=line,monitor=monitor,
                                            twiss=line.twiss(method='4d')),
                                            at_element='ps$start')


        Radius = np.sqrt(Monitor.x**2+Monitor.px**2)
        self.logger.debug(f"the shape of the radius matrix is {Radius.shape}")
        if plot:
            fig,ax = plt.subplots(2,1,figsize=(10,8),height_ratios=[2,1])
            ax[0].scatter(Monitor.x,Monitor.px)
            ax[0].set_aspects('equal')
            ax[1].plot(Radius[0,:]) # plotting only one particle?
        else :
            return Radius

    def add_kick_checks(self,start_at_turn=0,stop_at_turn=None,num_particles=None,**monitor_args):

        if stop_at_turn is None:
            stop_at_turn = self.sim_params['n_turns']
        if (num_particles is None)&('particle_id_range' not in monitor_args.keys()):
            num_particles= self.sim_params['n_part']

        #probably needed :
        # assert (num-particles is None('particle_id_range' not in monitor_args.keys())

        monitor_bf = xt.ParticlesMonitor(
            num_particles=num_particles,start_at_turn=start_at_turn,
            stop_at_turn=stop_at_turn,auto_to_numpy=True,**monitor_args)
        monitor_aft = xt.ParticlesMonitor(
            num_particles=num_particles,start_at_turn=start_at_turn,
            stop_at_turn=stop_at_turn,auto_to_numpy=True,**monitor_args)
        self.line.unfreeze()
        self.line.insert_element(element= monitor_bf,name='kickm_bf',index='drift_pr.kfb97..1')
        self.line.insert_element(element= monitor_aft,name='kickm_aft',index='drift_pr.kfb97..2')
        self.kick_monitors = [monitor_bf,monitor_aft]


import numpy as np
import matplotlib.pyplot as plt
import scipy
from tqdm import tqdm
import time as timing
import logging
logger = logging.getLogger(__name__)


import xtrack as xt
import xpart as xp

from .machine_class import Rfko


def nearest_betx(twiss_df, x, px, zeta):
    ''' Takes point in a transverse plane x-px and an array of longitudinal position in which
    translate those point along the line and add the effect of the closed orbit
    '''
    s = np.array(twiss_df.s)
    for ind in range(len(zeta)):
        nearest_point = s[np.argmin(np.abs(s-zeta[ind]))]
        x[ind] += twiss_df[twiss_df.s == nearest_point].x.iloc[0]
        px[ind] += twiss_df[twiss_df.s == nearest_point].px.iloc[0]

    return x, px

def nearest_betx2(twiss, x, px, zeta):
    ''' BETTER VERSION THAN ONE: Takes point in a transverse plane x-px and an array of longitudinal position in which
    translate those point along the line and add the effect of the closed orbit
    '''

    for i,z in enumerate(zeta):
        x[i] += np.interp(z,twiss.s,twiss.x)
        px[i] += np.interp(z,twiss.s,twiss.px)
    return x, px

def freq_range(outdict, plot=False):
    if isinstance(outdict,Rfko):
        outdict = outdict.to_dict()
    monitor = outdict['monitor']
    twiss = outdict['twiss']
    #twiss_df = twiss.to_pandas()

    freqs = 1 / twiss.T_rev0 - twiss.slip_factor * monitor.delta[monitor.delta != 0] * 1 / twiss.T_rev0
    if plot:
        plt.scatter(monitor.delta[monitor.delta != 0], freqs, s=2)
        plt.axhline(1 / twiss.T_rev0, linestyle='--', c='r')
        plt.title('Frequency range')
    return freqs


def tune_range(outdict, plot=False):
    if isinstance(outdict,Rfko):
        outdict = outdict.to_dict()
    monitor = outdict['monitor']
    twiss = outdict['twiss']
    #twiss_df = twiss.to_pandas()
    if (np.sum(monitor.delta)==0)|(outdict['params']['DPP_FACTOR']!=0):
        #It Means no simulation has been done still and the monitor didn't record any delta
        deltas = outdict['particles'].delta
    else:
        deltas = monitor.delta[monitor.delta != 0]
    tunes = twiss.qx + twiss.dqx *deltas
    if plot:
        plt.scatter(deltas, tunes, s=2)
        plt.axhline(twiss.qx, linestyle='--', c='r')
        plt.title('Tune range')
    return tunes



def find_multipole_component(line_df,order = 'sextupole'):
    ###### IT SEEMS THAT USING DATAFRAMES IS MUCH MORE EFFICIENT

    translation_dict_num = dict(dipole = 0, quadrupole = 1, sextupole = 2, octupole = 3)
    translation_dict_str = dict(dipole = 'Dipole', quadrupole = 'Quadrupole', sextupole = 'Sextupole', octupole = 'Octupole')
    if (order not in translation_dict_num.keys()) and (order not in translation_dict_num.values()) and \
            (order not in translation_dict_str.values()):
        print('Multipole component not recognized')
        return None
    ####### può sembrare un casino ma è piuttosto semplice- migliorarlo? magari con un unico dizionario??
    elif order in translation_dict_num.values():  # se dato come numero
        order_num = order
        order_str = list(translation_dict_num.keys())[
            list(translation_dict_num.values()).index(order)]  # to test for good
        order_corr = translation_dict_str[order_str]  # to have the correct available
    elif order in translation_dict_num.keys():  # se dato come stringa semplice
        order_num = translation_dict_num[order]
        order_str = order
        order_corr = translation_dict_str[order]
    else:  # altri casi già controllati tutti
        order_str = list(translation_dict_str.keys())[list(translation_dict_str.values()).index(order)]
        order_num = translation_dict_num[order_str]
        order_corr = order

    #print(f'order_str = {order_str}, order_num = {order_num}, order_corr = {order_corr}')

    ###### COMINCIA IL DIVERTIMENTO
    strength_attr = 'k'+str(order_num)
    # ELEMENTI DIRETTAMENTE INTERESSATI
    direct_names = line_df[line_df.element_type == order_corr].name # estrae elementi con ordine esplicito ed unico
    nz_direct_names = [i for i in direct_names if getattr(line_df[line_df.name == i].element.iloc[0],strength_attr) != 0]
    # ECCEZIONE QUADRUPOLI, GLI ELEMENTI CHE SONO QUADRUPOLI HANNO COMUNQUE DEFINITO ANCHE UN ARRAY DI COMPONENTI MULTIPOLARI
    quads = line_df[line_df.element_type == 'Quadrupole'].name
    if xt.__version__< '0.50.0' :
        quad_names = [i for i in quads if line_df[line_df.name == i].element.iloc[0]._order >= order_num]
    else:
        quad_names = []
        # in successive versions the quadrupole do not have multipole components,
        # but for compatibility with other functions I include an empty list


    nz_quad_names = [i for i in quad_names if line_df[line_df.name == i].element.iloc[0].knl[order_num]!= 0] # NB non zero anche per ordini diversi dai quadrupoli
    # COMBINED FUNCTION MAGNET
    combf = line_df[line_df.element_type == 'CombinedFunctionMagnet'].name
    combf_names = [i for i in combf if line_df[line_df.name == i].element.iloc[0]._order >= order_num]
    #combf_names seleziona solo quelli che comprendono l'ordine desiderato, per non incorrere il arrori di indice troopo lungo
    nz_combf_names = [i for i in combf_names if line_df[line_df.name == i].element.iloc[0].knl[order_num] != 0]
    # I COMB FUNCTION MAGNET HANNNO UN ALTRA ECCEZZIONE ECCEZZIONALE VERAMENTE : hanno singole componenti dipolo e quadrupolo
    if order_num in [0,1]:
        nz_combf_names_single = [i for i in combf if getattr(line_df[line_df.name == i].element.iloc[0],strength_attr)!=0]

    # MULTIPOLI
    mults = line_df[line_df.element_type == 'Multipole'].name
    mult_names = [i for i in mults if line_df[line_df.name == i].element.iloc[0]._order >= order_num]
    nz_mult_names = [i for i in mult_names if line_df[line_df.name == i].element.iloc[0].knl[order_num] !=0]
    if order_num in [0,1]:
        return dict(multipoles = nz_mult_names,combf_knl = nz_combf_names,quad_mult = nz_quad_names,direct_ele=nz_direct_names,combf_sing = nz_combf_names_single)
    return dict(multipoles = nz_mult_names,combf_knl = nz_combf_names,quad_mult = nz_quad_names,direct=nz_direct_names)



def Normalize_monitor(outdict,at_element='pe.smh57',keep_zeroes=True,center=True,remove_dispersion=True,normalize_particles=False):
    ''' TO ADD
    - Remove dispersion option
    - Particles norm?'''
    if isinstance(outdict,Rfko):
        rfko = outdict
        outdict = rfko.to_dict()

    monitor=outdict['monitor'].copy()
    try: # If the input is outdict referencng rfko raises an error
        if rfko.line_type=='henon_map':
            alfx, betx, _, _ = rfko.henon_params['twiss_params']
            monitor.x = rfko.monitor.x/np.sqrt(betx)
            monitor.px = monitor.x*alfx+ rfko.monitor.px*np.sqrt(betx)
            return monitor
    except:
        pass
    try:
        at_element = outdict['monitor'].placed_at_element
    except:
        pass

    logger.info(f'ELement considered for the normalization {at_element}')
    twiss = outdict['twiss']
    twiss_df = twiss.to_pandas()
    try:
        at_element = monitor.placed_at_element
        print('element considered based on PLACED AT ELEMENT')
    except:
        pass
    ######## CHECK WHAT KIND OF TRACKING HAS BEEN MADE
    if ('line_dynamics' in outdict.keys())|('dynamic_line' in outdict.keys()):
        if normalize_particles:
            logger.warning('dynamical normalization Norm_vary used')
            Monitor,Particles = Norm_vary(outdict,keep_zeroes=keep_zeroes, normalize_particles=normalize_particles,at_element=at_element,center=center,remove_dispersion=remove_dispersion)
            return Monitor,Particles
        else:
            logger.warning('dynamical normalization Norm_vary used')
            Monitor = Norm_vary(outdict,keep_zeroes=keep_zeroes, normalize_particles=normalize_particles,at_element=at_element,center=center)
            return Monitor


    else:
        # If no dynamical twiss is made optics stays constant hence betx,alfx....

        betx = twiss_df[twiss_df.name == at_element].betx.iloc[0]
        alfx = twiss_df[twiss_df.name == at_element].alfx.iloc[0]
        # Remove the orbit distortion and dispersion IF remove_dispersion is True
        if center:
            logger.debug(f'Remove dispersion option is = {remove_dispersion}')
            xc = twiss_df[twiss_df.name == at_element].x.iloc[0] + remove_dispersion*monitor.delta*twiss_df[twiss_df.name== at_element].dx.iloc[0]
            pxc = twiss_df[twiss_df.name == at_element].px.iloc[0]
        else:
            xc=0
            pxc=0

        if keep_zeroes:
            # If a monitor value is exactlty zero it means the particles has been extracted
            mask_null = np.array(monitor.x!=0 )
            X = (monitor.x - np.array(mask_null)*xc)/np.sqrt(betx)
            PX = X*alfx + (monitor.px-mask_null*pxc)*np.sqrt(betx)

        else:
            X = (monitor.x-xc)/np.sqrt(betx)
            PX = X*alfx + (monitor.px-pxc)*np.sqrt(betx)

        monitor.x = X
        monitor.px = PX
        if normalize_particles:
            logger.info('Particles normalization part is still unchecked')
            particles = outdict['particles'].copy()
            if center:
                X_p = (particles.x-xc)/np.sqrt(betx)
                PX_p = X_p*alfx + (particles.px-pxc)*np.sqrt(betx)
            else:
                X_p = (particles.x) / np.sqrt(betx)
                PX_p = X_p * alfx + (particles.px) * np.sqrt(betx)

            particles.x = X_p
            particles.px = PX_p
            return monitor,particles
        else:
            return monitor


def Norm_vary(outdict, keep_zeroes=True, normalize_particles=True, center=False, at_element='pe.smh57',remove_dispersion=True):
    ''' To correeeect absolutely the DISPERSION'''
    monitorc = outdict['monitor'].copy()
    monitor = outdict['monitor']
    partic = outdict['particles'].copy()
    n_turns = outdict['params']['n_turns']
    # out_dict['params']['dynamic_sextupole'] # in each case contains the number of sextupole ramping turns
    if 'twiss_dynamics' in outdict.keys():
        repeat_turns = n_turns
        xc = 0
        pxc = 0
        for n in range(repeat_turns):
            logger.debug('The normalization consider a fully dynamic situation using a different twiss every turn!')
            twiss_df = outdict['twiss_dynamics'][f'twiss_{n}'].to_pandas()
            betx = twiss_df[twiss_df.name == at_element].betx.iloc[0]
            alfx = twiss_df[twiss_df.name == at_element].alfx.iloc[0]
            # center to the distortion
            if center:
                logger.debug(f'Remove dispersion option is = {remove_dispersion}')
                xc = twiss_df[twiss_df.name == at_element].x.iloc[0]+ remove_dispersion*monitor.delta*twiss_df[twiss_df.name== at_element].dx.iloc[0]
                pxc = twiss_df[twiss_df.name == at_element].px.iloc[0]
            if keep_zeroes:
                mask_null = np.array(monitor.x[:, n] != 0)
                Xturn = (monitor.x[:, n] - np.array(mask_null) * xc) / np.sqrt(betx)
                PXturn = Xturn * alfx + (monitor.px[:, n] - mask_null * pxc) * np.sqrt(betx)
            else:
                Xturn = (monitor.x[:, n] - xc) / np.sqrt(betx)
                PXturn = Xturn * alfx + monitor.px[:, n] * np.sqrt(betx)
            if normalize_particles:
                logger.error('Particles normalization is not well working')
                particles = outdict['particles']
                X_pturn = (particles.x[:, n] - xc) / np.sqrt(betx)
                PX_pturn = X_pturn * alfx + (particles.px[:, n] - pxc) * np.sqrt(betx)
                partic.x[:, n] = X_pturn
                partic.px[:, n] = PX_pturn
            monitorc.x[:, n] = Xturn
            monitorc.px[:, n] = PXturn

    elif 'dynamic_twiss' in outdict.keys():
        logger.debug('The normalization consider the sextupole ramping!')
        repeat_turns = n_turns
        ramping_stop = outdict['params']['dynamic_sextupole']
        done_flag = True  # Turn false when the action is done
        xc = 0
        pxc = 0
        twiss_df0 = outdict['twiss'].to_pandas()
        betx0 = twiss_df0[twiss_df0.name == at_element].betx.iloc[0]
        alfx0 = twiss_df0[twiss_df0.name == at_element].alfx.iloc[0]
        dx0 =  twiss_df0[twiss_df0.name == at_element].dx.iloc[0]

        for n in range(repeat_turns):
            if n + 1 <= ramping_stop:
                twiss_df = outdict['dynamic_twiss'][f'twiss_{n + 1}'].to_pandas()
                betx = twiss_df[twiss_df.name == at_element].betx.iloc[0]
                alfx = twiss_df[twiss_df.name == at_element].alfx.iloc[0]
                dx = remove_dispersion*monitor.delta*twiss_df[twiss_df.name== at_element].dx.iloc[0]
            elif done_flag:
                twiss_df = twiss_df0
                betx = betx0
                alfx = alfx0
                dx = dx0
                # ACTION FLAG TO NOT REPEAT EVERYTIME
                done_flag = False

                # center to the distortion
            if center:
                logger.debug(f'Remove dispersion option is : {remove_dispersion}')
                xc = twiss_df[twiss_df.name == at_element].x.iloc[0] + remove_dispersion*monitor.delta[:, n]*dx
                pxc = twiss_df[twiss_df.name == at_element].px.iloc[0]
            if keep_zeroes:
                mask_null = np.array(monitor.x[:, n] != 0)
                Xturn = (monitor.x[:, n] - np.array(mask_null) * xc) / np.sqrt(betx)
                PXturn = Xturn * alfx + (monitor.px[:, n] - mask_null * pxc) * np.sqrt(betx)
            else:
                Xturn = (monitor.x[:, n] - xc) / np.sqrt(betx)
                PXturn = Xturn * alfx + monitor.px[:, n] * np.sqrt(betx)
            if normalize_particles:
                logger.error('The normalization of particles is not correct')
                particles = outdict['particles']
                X_pturn = (particles.x[:, n] - xc) / np.sqrt(betx)
                PX_pturn = X_pturn * alfx + (particles.px[:, n] - pxc) * np.sqrt(betx)
                partic.x[:, n] = X_pturn
                partic.px[:, n] = PX_pturn
            monitorc.x[:, n] = Xturn
            monitorc.px[:, n] = PXturn

    if normalize_particles:
        return monitorc, partic
    else:
        return monitorc






def norm_part(outdict,particles=None, at_element=None,at_s=None):
    ''' Could consider that is noither at_element nor at_s are given I could read the position of the particles'''
    if isinstance(outdict,Rfko):
        outdict = outdict.to_dict(particles=particles)
    assert (at_element is not None)|(at_s is not None), 'Give at_s or at_element'
    assert not (at_element is None)&(at_s is None), 'Give at_s OR at_element' # check description
    if particles is None:
        particles = outdict['particles']
    twiss_df = outdict['twiss'].to_pandas()
    Particles = particles.copy()
    if at_element is not None: # if generated at the same locations
        betx = twiss_df[twiss_df.name == at_element].betx.iloc[0]
        alfx = twiss_df[twiss_df.name == at_element].alfx.iloc[0]
        xc = twiss_df[twiss_df.name == at_element].x.iloc[0]
        pxc = twiss_df[twiss_df.name == at_element].px.iloc[0]
        Particles.x = (particles.x - xc) / np.sqrt(betx)
        Particles.px = Particles.x * alfx + (particles.px - pxc) * np.sqrt(betx)
    else:  # When generated at different amplitudes
        xc = np.interp(at_s, twiss_df.s, twiss_df.x)
        pxc = np.interp(at_s, twiss_df.s, twiss_df.px)
        betx = np.interp(at_s, twiss_df.s, twiss_df.betx)
        alfx = np.interp(at_s, twiss_df.s, twiss_df.alfx)
        Particles.x = (particles.x - xc) / np.sqrt(betx)
        Particles.px = Particles.x * alfx + (particles.px - pxc) * np.sqrt(betx)


    return Particles


def DEnormalize_points(rfko, X=None, PX=None, at_element=None, at_s=None, centered=True):
    '''inverse transform, to denormalize'''
    assert (at_element is not None) | (at_s is not None), 'Choose either at_element or at_s but not both'

    twiss = rfko.twiss
    ref_s = [at_s if (at_s is not None) else twiss.s[twiss.name == at_element]]
    if at_element is not None:
        betx = twiss.betx[twiss.name == at_element]
        alfx = twiss.alfx[twiss.name == at_element]
        xc = twiss.x[twiss.name == at_element]
        pxc = twiss.px[twiss.name == at_element]
    else:
        betx = np.interp(at_s, twiss.s, twiss.betx)
        alfx = np.interp(at_s, twiss.s, twiss.alfx)
        xc = np.interp(at_s, twiss.s, twiss.x)
        pxc = np.interp(at_s, twiss.s, twiss.px)

    if not centered:
        xc = 0
        pxc = 0
    x = X * np.sqrt(betx) + xc
    px = (PX - X * alfx) / np.sqrt(betx) + pxc
    return x, px



def tracking(out_dict, type='first', sextupoles_turns=None, twiss=True,return_time=False):
    ''' Tracking of a line with the outdict convention, 3 different type of tracking

    CHANGES/IMPOROVEMENT NECESSARY
    - third type: turning off all the extraction elements is better (quads)
    - different type (?): to do activate the extraction sextupole after a while


    :param twiss:
    :param out_dict:
    :param type:
    :param sextupoles_turns:
    :return:
    '''
    if isinstance(out_dict,Rfko):
        rfko = out_dict
        out_dict = out_dict.to_dict()

    line = out_dict['line']
    particles = out_dict['particles']
    if line.tracker is None:
        line.build_tracker()
    if type == 'second':
        twiss_dict = dict()
        line_dict = dict()
    # Needed to know the turns
    try:
        n_turns = out_dict['params']['n_turns']
    except:
        print('turns taken from the first particle')
        n_turns = len(out_dict['monitor'].x[0, :])

    if type == 'first':
        print('Extracting sextupoles already active')
        # Tracking 1
        time_before = timing.perf_counter()
        print("> Tracking...")
        for _ in tqdm(range(n_turns)):
            line.track(particles, num_turns=1)
        time_after = timing.perf_counter()
        time_taken = time_after - time_before
        print(f"> Tracking done, {time_taken} seconds elapsed")

        particles_lost_at_each_turn = np.bincount(particles.at_turn)
        print(f'total extracted particles :{np.sum(particles_lost_at_each_turn[:-1])} \n')
        print("! Tracking Complete !")
        if return_time:
            return time_taken

    elif type == 'second':
        #  turns to activate completely the sextupoles
        if sextupoles_turns == None:
            print('turns to activate sextupoles non defined, considering 100 turns')
            sextupoles_turns = 100
        out_dict['params']['dynamic_sextupole'] = sextupoles_turns  #
        xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
        setter = xt.MultiSetter(line, xse_names, field='k2')  # multisetter obj
        xse0 = setter.get_values()  # initial/nominal values
        it = 0  # counter for the ramp
        ramp = np.linspace(0, 1, sextupoles_turns)
        # 1)Tracking-ramping sextupoles
        time_before = timing.perf_counter()
        print("> Sextupoles adiabatic turning on...")

        for _ in tqdm(range(sextupoles_turns)):
            k2_it = xse0 * ramp[it]
            setter.set_values(k2_it)  # changes of value
            if twiss:
                twiss_dict[f'twiss_{it + 1}'] = line.twiss(method='4d')
            line_dict[f'line_{it + 1}'] = line
            line.track(particles, num_turns=1)
            it += 1
        time_after = timing.perf_counter()
        time_taken = time_after - time_before
        print(f"> sextupoles at nominal value, {time_taken} seconds elapsed")
        if twiss:
            print('calculating twiss')
            out_dict['dynamic_twiss'] = twiss_dict
        out_dict['dynamic_line'] = line_dict

        # 2)Tracking at nominal values
        for _ in tqdm(range(n_turns - sextupoles_turns)):
            line.track(particles, num_turns=1)
        time_after2 = timing.perf_counter()
        time_taken2 = time_after2 - time_before
        print(f"> tracking finished, total time: {time_taken2} seconds")
        particles_lost_at_each_turn = np.bincount(particles.at_turn)
        print(f'total extracted particles :{np.sum(particles_lost_at_each_turn[:-1])} \n')
        if return_time:
            return time_taken+time_taken2


    elif type == 'third':  # EXTRACTION SEXTUPOLES COMPLETELY OFF
        xse_names = ['pr.xse01.a', 'pr.xse01.b', 'pr.xse07']
        # SHOULD TURN OFF ALSO THE EXTRACTION QUADRUPOLEs?!
        setter = xt.MultiSetter(line, xse_names, field='k2')  # multisetter obj
        xse0 = setter.get_values()  # initial/nominal values
        setter.set_values(xse0 * 0)  # set to 0
        ### SHOULD BE RESET TO THE INITIAL VALUE

        # if out:
        #     print('extraction sextupoles off, twiss in outdict replaced')
        #     outdict['twiss'] = line.twiss(method='4d')

        time_before = timing.perf_counter()
        for _ in tqdm(range(n_turns)):
            line.track(particles, num_turns=1)
        time_taken = timing.perf_counter() - time_before
        print(f"> tracking finishes, total time: {time_taken} seconds")
        particles_lost_at_each_turn = np.bincount(particles.at_turn)
        print(f'total extracted particles :{np.sum(particles_lost_at_each_turn[:-1])} \n')
        if return_time:
            return time_taken

    else:  ## FOR A WRONG ENTRY OF THE TYPE PARAMETER
        print('Tracking type not recognized')


def build_particles(line, n_part=1000, DPP_FACTOR=2.35e-3, exn=1.5e-6, eyn=1.5e-6, twiss=None, out_params=False):
    # Bunched is an option that for now I am not using
    if out_params:
        params = locals()
    if twiss is None:
        twiss = line.twiss(method='4d')

    sigmas, sigmas_p = xp.generate_2D_gaussian(int(n_part))

    particles = line.build_particles(
        method='4d',
        delta=DPP_FACTOR,
        nemitt_x=exn,
        nemitt_y=eyn,
        x_norm=sigmas,
        px_norm=sigmas_p,
        y_norm=sigmas,
        py_norm=sigmas_p,
    )

    # Debunched creation in the z plane
    z_array = scipy.stats.uniform.rvs(loc=0 - twiss['circumference'] / 2,
                                      scale=twiss['circumference'],
                                      size=int(n_part))
    delta_array = scipy.stats.norm.rvs(loc=0, scale=DPP_FACTOR / 2, size=int(n_part))
    particles.zeta = z_array
    particles.delta = delta_array
    if out_params:
        return particles, params
    else:
        return particles
